import os

import mne
import numpy as np
import pytest
from bids import BIDSLayout
from mne.io import BaseRaw
from mne_bids import BIDSPath

from ieeg.io import raw_from_layout

bids_root = mne.datasets.epilepsy_ecog.data_path()
seeg = mne.io.read_raw(mne.datasets.misc.data_path() /
                       'seeg' / 'sample_seeg_ieeg.fif')
layout = BIDSLayout(bids_root)


def test_bids():
    assert "pt1" in layout.get_subjects()


def test_bidspath_from_layout():
    from ieeg.io import bidspath_from_layout
    expected = "sub-pt1_ses-presurgery_task-ictal_ieeg.eeg"
    bidspath = bidspath_from_layout(layout, subject="pt1",
                                    extension=".eeg")
    assert isinstance(bidspath, BIDSPath)
    assert bidspath.basename == expected


def test_raw_from_layout():
    raw = raw_from_layout(layout, subject="pt1", extension=".vhdr")
    assert isinstance(raw, BaseRaw)


@pytest.mark.parametrize("n_jobs", [1, 8])
def test_line_filter(n_jobs):
    from ieeg.mt_filter import line_filter
    mne.set_log_file("output.log",
                     "%(levelname)s: %(message)s - %(asctime)s",
                     overwrite=True)
    mne.set_log_level("DEBUG")
    raw = raw_from_layout(layout, subject="pt1", preload=True,
                          extension=".vhdr")
    filt = line_filter(raw, raw.info['sfreq'], [60])
    raw_dat = raw._data
    filt_dat = filt._data
    assert filt_dat.shape == raw_dat.shape
    params = dict(method='multitaper', tmax=20, fmin=55, fmax=65,
                  bandwidth=0.5, n_jobs=n_jobs)
    rpsd = raw.compute_psd(**params)
    fpsd = filt.compute_psd(**params)
    assert np.mean(np.abs(rpsd.get_data() - fpsd.get_data())) > 1e-10


if os.path.isfile("spec.npy"):
    spec_check = np.load("spec.npy")
else:
    spec_check = np.load("tests/spec.npy")


def test_spect_1():
    from ieeg.timefreq.multitaper import spectrogram
    raw = raw_from_layout(layout, subject="pt1", preload=True,
                          extension=".vhdr")
    freqs = np.arange(10, 20, 2)
    spectra = spectrogram(raw, freqs, 'onset', -0.5, 0.5, 'onset', -1, -0.5,
                          n_jobs=-1, picks=[0, 1], decim=10, pad="0.5s")
    out = spectra._data
    assert np.allclose(out, spec_check)


def test_spect_2():
    from ieeg.timefreq.multitaper import spectrogram
    from ieeg.navigate import trial_ieeg
    raw = raw_from_layout(layout, subject="pt1", preload=True,
                          extension=".vhdr")
    freqs = np.arange(10, 20, 2)
    on1 = trial_ieeg(raw, 'onset', (-1, 1))
    on2 = trial_ieeg(raw, 'onset', (-1.5, 0))
    spectra = spectrogram(on1, freqs, on2, n_jobs=-1, picks=[0, 1], decim=10,
                          pad="0.5s")
    out = spectra._data
    assert np.allclose(out, spec_check)


@pytest.mark.parametrize("input1, input2, expected", [
    (4, np.inf, ['LAMY 7', 'RAHP 3']),
    (3, 2, ['LAMY 7', 'LPHG 6', 'LBRI 3', 'RAHP 3', 'LENT 3', 'LPIT 5'])
])
def test_outlier(input1, input2, expected):
    from ieeg.navigate import channel_outlier_marker
    outs = channel_outlier_marker(seeg, input1, input2)
    assert outs == expected


@pytest.mark.parametrize("outliers, n_out", [
    (4, 89955),
    (5, 25987)
])
def test_trial_outlier(outliers, n_out):
    from ieeg.navigate import trial_ieeg, outliers_to_nan
    trials = trial_ieeg(seeg, 'Response', (-1, 1))
    outs = outliers_to_nan(trials, outliers)
    assert np.isnan(outs._data).sum() == n_out



import argparse
from typing import Union

import numpy as np
from mne.io import pick
from mne.utils import fill_doc, logger, verbose

from ieeg import ListNum
from ieeg.process import proc_array
from ieeg.timefreq import utils as mt_utils
from ieeg.timefreq.multitaper import WindowingRemover


@fill_doc
@verbose
def line_filter(raw: mt_utils.Signal, fs: float = None, freqs: ListNum = 60.,
                filter_length: str = '10s', notch_widths: ListNum = 10.,
                mt_bandwidth: float = None, p_value: float = 0.05,
                picks: list[Union[int, str]] = None, n_jobs: int = None,
                adaptive: bool = True, low_bias: bool = True,
                copy: bool = True, *, verbose: Union[int, bool, str] = None
                ) -> mt_utils.Signal:
    """Apply a multitaper line noise notch filter for the signal instance.

    Applies a multitaper power line noise notch filter to the signal, operating
    on the last dimension. Uses the F-test to find significant sinusoidal
    components to remove. This is done by fitting a sinusoid to the power
    spectrum at each time point and frequency, and testing whether the
    resulting fit is significantly different from a flat spectrum. The
    significance test is done using an F-test, which requires fitting two
    models (one flat, one sinusoidal) at each time point and frequency. The
    F-test is corrected for multiple comparisons using the Benjamini-Hochberg
    procedure.

    Parameters
    ----------
    raw : mt_utils.Signal
        Signal to filter.
    fs : float, optional
        Sampling rate in Hz. Default is taken from the raw object.
    freqs : float | array-like of float, optional
        Frequencies to notch filter in Hz, e.g. np.arange(60, 241, 60).
        None can only be used with the mode 'spectrum_fit', where an F
        test is used to find sinusoidal components.
    filter_length : str | int, optional
        Length of the filter to use. If str, assumed to be human-readable time
        in units of "s" or "ms" (e.g., "10s" or "5500ms"). If an int, it
        is assumed to be in samples and used directly.
    notch_widths : float | array of float, optional
        Width of the stop band (centred at each freq in freqs) in Hz.
        Default is 10.
    mt_bandwidth : float, optional
        The bandwidth of the multitaper windowing function in Hz. Default will
        set the half frequency bathwidth to 4 Hz.
    p_value : float, optional
        P-value to use in F-test thresholding to determine significant
        sinusoidal components to remove. Note that this will be Bonferroni
        corrected for the number of frequencies, so large p-values may be
        justified.
    %(picks_all)s
    %(n_jobs)s
    adaptive : bool, optional
        Use adaptive weights to combine the tapered spectra into PSD.
        Default is True.
    low_bias : bool, optional
        Only use tapers with more than 90 percent spectral concentration within
        bandwidth. Default is True.
    copy : bool, optional
        If True, a copy of x, filtered, is returned. Otherwise, it operates
        on x in place.
    %(verbose)s

    Returns
    -------
    filt : mt_utils.Signal
        The signal instance with the filtered data.

    See Also
    --------
    <https://mne.tools/stable/generated/mne.filter.notch_filter.html>

    Notes
    -----
    The frequency response is (approximately) given by
    ::

        1-|----------         -----------
          |          \\       /
          |           \\     /
          |            \\   /
          |             \\ /
        0-|              -
          |         |    |    |         |
          0        Fp1 freq  Fp2       Nyq

    For each freq in freqs, where ``Fp1 = freq - trans_bandwidth / 2`` and
    ``Fs2 = freq + trans_bandwidth / 2``.

    References
    ----------
    Multi-taper removal is inspired by code from the Chronux toolbox, see
    www.chronux.org and the book "Observed Brain Dynamics" by Partha Mitra
    & Hemant Bokil, Oxford University Press, New York, 2008. Please
    cite this in publications if this function is used.

    Examples
    --------
    >>> import mne
    >>> from bids import BIDSLayout
    >>> from ieeg.io import raw_from_layout
    >>> bids_root = mne.datasets.epilepsy_ecog.data_path(verbose=False)
    >>> layout = BIDSLayout(bids_root)
    >>> raw = raw_from_layout(layout, subject="pt1", preload=True,
    ... extension=".vhdr", verbose=False)
    Reading 0 ... 269079  =      0.000 ...   269.079 secs...
    >>> mne.set_log_level("WARNING")
    >>> filt = line_filter(raw, freqs=[60, 120, 180])
    >>> mne.set_log_level("INFO")
    """
    if fs is None:
        fs = raw.info["sfreq"]
    if copy:
        filt = raw.copy()
    else:
        filt = raw

    x = filt.get_data("data")
    x = mt_utils._check_filterable(x, 'notch filtered', 'notch_filter')
    if freqs is not None:
        freqs = np.atleast_1d(freqs)
        # Only have to deal with notch_widths for non-autodetect
        if notch_widths is None:
            notch_widths = freqs / 200.0
        elif np.any(notch_widths < 0):
            raise ValueError('notch_widths must be >= 0')
        else:
            notch_widths = np.atleast_1d(notch_widths)
            if len(notch_widths) == 1:
                notch_widths = notch_widths[0] * np.ones_like(freqs)
            elif len(notch_widths) != len(freqs):
                raise ValueError('notch_widths must be None, scalar, or the '
                                 'same length as freqs')

    data_idx = np.where([ch_t in set(raw.get_channel_types(
        only_data_chs=True)) for ch_t in raw.get_channel_types()])[0]

    # convert filter length to samples
    if filter_length is None:
        filter_length = x.shape[-1]

    filter_length: int = min(mt_utils.to_samples(filter_length, fs),
                             x.shape[-1])

    process = WindowingRemover(fs, freqs, notch_widths, filter_length,
                               adaptive, low_bias, mt_bandwidth, p_value)

    filt._data[data_idx] = mt_spectrum_proc(x, process, picks, n_jobs)

    return filt


def mt_spectrum_proc(x: np.ndarray, process: callable, picks: list,
                     n_jobs: int) -> np.ndarray:
    """Call _mt_spectrum_remove."""
    # set up array for filtering, reshape to 2D, operate on last axis
    x, orig_shape, picks = _prep_for_filtering(x, picks)

    proc_array(process, x, n_jobs=n_jobs, desc="Channels")

    x.shape = orig_shape
    return x


def _prep_for_filtering(x: np.ndarray, picks: list = None
                        ) -> tuple[np.ndarray, tuple, int]:
    """Set up array as 2D for filtering ease."""
    x = mt_utils._check_filterable(x)
    orig_shape = x.shape
    x = np.atleast_2d(x)
    picks = pick._picks_to_idx(x.shape[-2], picks)
    x.shape = (np.prod(x.shape[:-1]), x.shape[-1])
    if len(orig_shape) == 3:
        n_epochs, n_channels, n_times = orig_shape
        offset = np.repeat(np.arange(0, n_channels * n_epochs, n_channels),
                           len(picks))
        picks = np.tile(picks, n_epochs) + offset
    elif len(orig_shape) > 3:
        raise ValueError('picks argument is not supported for data with more'
                         ' than three dimensions')
    assert all(0 <= pick < x.shape[0] for pick in picks)  # guaranteed by above

    return x, orig_shape, picks


def _get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter, description="""
            """,
        epilog="""
            Made by Aaron Earle-Richardson (ae166@duke.edu)
            """)

    parser.add_argument("-s", "--subject", required=False, default=None,
                        help="data subject to clean")
    return parser


def _main(subject: str = None, save: bool = False):
    import mne
    from bids import BIDSLayout
    from ieeg.io import raw_from_layout, save_derivative
    import os

    HOME = os.path.expanduser("~")
    LAB_root = os.path.join(HOME, "Box", "CoganLab")
    # %% Set up logging
    mne.set_log_file("output.log",
                     "%(levelname)s: %(message)s - %(asctime)s",
                     overwrite=True)
    mne.set_log_level("INFO")

    bids_root = LAB_root + "/BIDS-1.3_SentenceRep/BIDS"
    layout = BIDSLayout(bids_root)
    if subject is not None:
        do_subj = [subject]
    else:
        do_subj = layout.get(return_type="id", target="subject")
    do_subj.sort()
    if 'SLURM_ARRAY_TASK_ID' in os.environ.keys():
        taskIDs = [int(os.environ['SLURM_ARRAY_TASK_ID'])]
    else:
        taskIDs = list(range(len(do_subj)))
    for id in taskIDs:
        subj = do_subj[id]
        try:
            raw = raw_from_layout(layout, subject=subj, extension=".edf",
                                  preload=False)
            # %% filter data
            filt = line_filter(raw, mt_bandwidth=10., n_jobs=-1,
                               filter_length='700ms', verbose=10,
                               freqs=[60], notch_widths=20)
            filt2 = line_filter(filt, mt_bandwidth=10., n_jobs=-1,
                                filter_length='20s', verbose=10,
                                freqs=[60, 120, 180, 240], notch_widths=20)
            # %% Save the data
            if save:
                save_derivative(filt2, layout, "clean")
        except Exception as e:
            logger.error(e)


if __name__ == "__main__":
    args = _get_parser().parse_args()
    _main(**vars(args))



from functools import singledispatch

import numpy as np
from mne import Epochs
from mne.io import Raw, base
from tqdm import tqdm
from joblib import Parallel, delayed

from ieeg.process import COLA, cpu_count, get_mem, parallelize
from ieeg.timefreq.utils import BaseEpochs, Evoked, Signal
from ieeg.timefreq.hilbert import (filterbank_hilbert_first_half_wrapper,
                                   extract_channel_wrapper)


@singledispatch
def extract(data: np.ndarray, fs: int = None,
            passband: tuple[int, int] = (70, 150), copy: bool = True,
            n_jobs=-1, verbose: bool = True) -> np.ndarray:
    """Extract gamma band envelope from data.

    Parameters
    ----------
    data : (np.ndarray, shape ((epochs) ,channels, samples)) | Signal
        Data to extract gamma envelope from. If Signal, will use the _data
        attribute.
    fs : int, optional
        Sampling frequency of data. If Signal, will use the data.info['sfreq'].
        Otherwise, must be provided.
    passband : tuple[int, int], optional
        Passband in Hz, high gamma band by default (70, 150)
    copy : bool, optional
        Whether to copy data or operate in place if False, by default True
    n_jobs : int, optional
        Number of jobs to run in parallel, by default all available cores

    Returns
    -------
    np.ndarray
        Gamma envelope.

    Notes
    -----
    This function is a wrapper for
    `filterbank_hilbert <https://naplib-python.readthedocs.io/en/latest/referen
    ces/preprocessing.html#naplib.preprocessing.filterbank_hilbert>`_. It is a
    convenience function for extracting the gamma envelope from data. It is
    optimized for speed, but not memory. If you have a lot of data, you may
    want to epoch your data first and then extract the envelope.

    Examples
    --------
    >>> import mne
    >>> import numpy as np
    >>> from bids import BIDSLayout
    >>> from ieeg.navigate import trial_ieeg
    >>> from ieeg.io import raw_from_layout
    >>> from ieeg.timefreq.utils import crop_pad
    >>> bids_root = mne.datasets.epilepsy_ecog.data_path(verbose=50)
    >>> layout = BIDSLayout(bids_root)
    >>> raw = raw_from_layout(layout, subject="pt1",
    ... preload=True, extension=".vhdr", verbose=False) # doctest: +ELLIPSIS
    Reading 0 ... 269079  =      0.000 ...   269.079 secs...
    >>> trials = trial_ieeg(raw, "AD1-4, ATT1,2", (-0.5, 1),
    ... preload=True, verbose=False, picks=['AD2'])
    >>> gamma = extract(trials, n_jobs=1)
    >>> crop_pad(gamma, "0.5s") # doctest: +ELLIPSIS
    <Epochs |  1 events (all good), 0 – 0.5 s, baseline off, ...
     'AD1-4, ATT1,2': 1>
    >>> gamma.resample(100, verbose=50) # doctest: +ELLIPSIS
    <Epochs |  1 events (all good), 0 – 0.49 s, baseline off, ...
     'AD1-4, ATT1,2': 1>
    >>> expected = np.array([
    ... 3.5729, 3.8323, 4.0820, 5.4100, 8.0623, 12.579, 20.280, 31.027, 43.918,
    ... 56.523, 65.739, 68.678, 64.378, 54.357, 42.245, 32.177, 26.568, 23.578,
    ... 20.584, 17.003, 13.105, 9.6693, 6.9391, 4.8671, 3.6392, 3.0246, 2.8268,
    ... 2.9109, 3.2376, 3.6906, 4.1659, 4.5842, 4.9949, 5.3240, 6.0321, 7.1968,
    ... 8.0531, 8.4710, 8.3094, 7.8219, 7.3717, 7.1496, 7.0281, 7.0632, 7.0525,
    ... 7.2997, 7.7566, 7.7874, 7.3208, 6.4729]) * 1e-05
    >>> np.abs(np.sum(gamma._data - expected)) < 1e-6
    True
    """

    if fs is None:
        raise ValueError("fs must be provided if data is not a Signal")

    if copy:
        in_data = data.copy()
    else:
        in_data = data

    passband = list(passband)
    env = np.zeros(in_data.shape)

    if len(in_data.shape) == 3:  # Assume shape is (trials, channels, time)
        trials = range(in_data.shape[0])
        if n_jobs != 1:
            ins = (in_data[trial].T for trial in trials)
            par_out = parallelize(filterbank_hilbert, ins, fs=fs, Wn=passband,
                                  n_jobs=n_jobs)
            env[:, :, :] = np.array([np.sum(out, axis=-1).T for
                                     out in par_out])
        else:
            if verbose:
                trials = tqdm(trials)
            for trial in trials:
                out = filterbank_hilbert(in_data[trial, :, :].T, fs,
                                         passband, 1)
                env[trial, :, :] = np.sum(out, axis=-1).T
    elif len(in_data.shape) == 2:  # Assume shape is (channels, time)
        out = filterbank_hilbert(in_data.T, fs, passband, n_jobs)
        env = np.sum(out, axis=-1).T
    else:
        raise ValueError("number of dims should be either 2 or 3, not {}"
                         "".format(len(in_data.shape)))

    return env


def _extract_inst(inst: Signal, fs: int, copy: bool, **kwargs) -> Signal:
    if fs is None:
        fs = inst.info['sfreq']
    if copy:
        sig = inst.copy()
    else:
        sig = inst

    sig._data = extract(sig._data, fs, copy=False, **kwargs)

    return sig


@extract.register
def _(inst: base.BaseRaw, fs: int = None,
      passband: tuple[int, int] = (70, 150),
      copy: bool = True, n_jobs=-1, verbose: bool = True) -> Raw:
    """Extract gamma band envelope from Raw object."""
    return _extract_inst(inst, fs, copy, passband=passband, n_jobs=n_jobs,
                         verbose=verbose)


@extract.register
def _(inst: BaseEpochs, fs: int = None,
      passband: tuple[int, int] = (70, 150),
      copy: bool = True, n_jobs=-1, verbose: bool = True) -> Epochs:
    """Extract gamma band envelope from Epochs object."""
    return _extract_inst(inst, fs, copy, passband=passband, n_jobs=n_jobs,
                         verbose=verbose)


@extract.register
def _(inst: Evoked, fs: int = None,
      passband: tuple[int, int] = (70, 150),
      copy: bool = True, n_jobs=-1, verbose: bool = True) -> Evoked:
    """Extract gamma band envelope from Evoked object."""
    return _extract_inst(inst, fs, copy, passband=passband, n_jobs=n_jobs,
                         verbose=verbose)


def _my_hilt(x: np.ndarray, fs, Wn=(1, 150), n_jobs=-1):
    # Set default window function and threshold
    cfs = get_centers(Wn)
    n_times = x.shape[0]
    bytes_per = 16  # numpy complex128 is 16 bytes per num
    chunk_size = min([x.size * bytes_per * len(cfs), get_mem()])
    n_samples = int(chunk_size / (cpu_count() * x.shape[1] * len(cfs)))
    n_overlap = (n_samples + 1) // 2
    x_out = np.zeros_like(x.T)
    idx = [0]

    # Define how to process a chunk of data
    def process(x_):
        out = filterbank_hilbert(x_, fs, Wn, 1)
        return (out[1])  # must return a tuple

    # Define how to store a chunk of fully processed data (it's trivial)
    def store(x_):
        stop = idx[0] + x_.T.shape[-1]
        x_out[..., idx[0]:stop] += x_.T
        idx[0] = stop

    COLA(process, store, n_times, n_samples, n_overlap, fs,
         n_jobs=n_jobs).feed(x)
    assert idx[0] == n_times
    return x_out, cfs


def get_centers(Wn):
    """Get center frequencies for filter bank.

    Parameters
    ----------
    Wn : tuple
        Frequency range to use for filter bank.

    Returns
    -------
    cfs : list
        Center frequencies for filter bank.
        """

    # create filter bank
    a = np.array([np.log10(0.39), 0.5])
    f0 = 0.018
    octSpace = 1. / 7
    minf, maxf = Wn
    if minf >= maxf:
        raise ValueError(
            f'Upper bound of frequency range must be greater than lower '
            f'bound, but got lower bound of {minf} and upper bound of {maxf}')
    maxfo = np.log2(maxf / f0)  # octave of max freq

    cfs = [f0]
    sigma_f = 10 ** (a[0] + a[1] * np.log10(cfs[-1]))

    while np.log2(cfs[-1] / f0) < maxfo:

        if cfs[-1] < 4:
            cfs.append(cfs[-1] + sigma_f)
        else:  # switches to log spacing at 4 Hz
            cfo = np.log2(cfs[-1] / f0)  # current freq octave
            cfo += octSpace  # new freq octave
            cfs.append(f0 * (2 ** (cfo)))

        sigma_f = 10 ** (a[0] + a[1] * np.log10(cfs[-1]))

    cfs = np.array(cfs)
    if np.logical_and(cfs >= minf, cfs <= maxf).sum() == 0:
        raise ValueError(
            f'Frequency band is too narrow, so no filters in filterbank are '
            f'placed inside. Try a wider frequency band.')

    cfs = cfs[np.logical_and(cfs >= minf,
                             cfs <= maxf)]  # choose those that lie in the
    # input freqRange
    return cfs


def filterbank_hilbert(x, fs, Wn=[70, 150], n_jobs=1):
    """
    Compute the phase and amplitude (envelope) of a signal for a single
    frequency band, as in [#edwards]_. This is done using a filter bank of
    gaussian shaped filters with center frequencies linearly spaced until 4Hz
    and then logarithmically spaced. The Hilbert Transform of each filter's
    output is computed and the amplitude and phase are computed from the
    complex values. See [#edwards]_ for details on the filter bank used.

    See Also
    --------
    filter_hilbert

    Parameters
    ----------
    x : np.ndarray, shape (time, channels)
        Signal to filter. Filtering is performed on each channel independently.
    fs : int
        Sampling rate.
    Wn : list or array-like, length 2, default=[70, 150]
        Lower and upper boundaries for filterbank center frequencies. A range
        of [1, 150] results in 42 filters.
    n_jobs : int, default=1
        Number of jobs to use to compute filterbank across channels in
        parallel.

    Returns
    -------
    x_phase : np.ndarray, shape (time, channels, frequency_bins)
        Phase of each frequency bin in the filter bank for each channel.
    x_envelope : np.ndarray, shape (time, channels, frequency_bins)
        Envelope of each frequency bin in the filter bank for each channel.
    center_freqs : np.ndarray, shape (frequency_bins,)
        Center frequencies for each frequency bin used in the filter bank.
    Examples
    --------
    >>> import numpy as np
    >>> x = np.random.rand(1000,3) # 3 channels of signals
    >>> fs = 500
    >>> x_envelope = filterbank_hilbert(x, fs, Wn=[1, 150])
    >>> # the outputs have the phase and envelope for each channel and each
    filter in the filterbank
    >>> x_envelope.shape # 3rd dimension is one for each filter in filterbank
    (1000, 3, 42)

    """

    x = x.astype('float32')
    minf, maxf = Wn

    if minf >= maxf:
        raise ValueError(
            (f'Upper bound of frequency range must be greater than lower bound'
             f', but got lower bound of {minf} and upper bound of {maxf}'))

    Xf, freqs, cfs, N, sds, h = filterbank_hilbert_first_half_wrapper(
        x, fs, minf, maxf)

    def extract_channel(Xf):
        return extract_channel_wrapper(Xf, freqs, cfs, N, sds, h, minf, maxf)

    # pre-allocate
    hilb_amp = np.zeros((*x.shape, len(cfs)), dtype='float32')

    # process channels sequentially
    if n_jobs == 1:
        for chn in range(x.shape[1]):
            hilb_amp[:, chn] = extract_channel(Xf[:, chn])
    # process channels in parallel
    else:
        results = Parallel(n_jobs)(delayed(extract_channel)(
            Xf[:, chn]) for chn in range(x.shape[1]))
        for chn, amp in enumerate(results):
            hilb_amp[:, chn] = amp

    return hilb_amp



# %% IMPORT PACKAGES

import numpy as np
from numpy.linalg import inv as inv  # Used in kalman filter
from scipy.spatial.distance import pdist
from scipy.spatial.distance import squareform
from scipy.stats import norm
from scipy.spatial.distance import cdist
import math
from sklearn.pipeline import Pipeline
from sklearn import linear_model  # For Wiener Filter and Wiener Cascade
from sklearn.svm import SVR  # For support vector regression (SVR)
from sklearn.svm import SVC  # For support vector classification (SVM)
from sklearn.decomposition import PCA  # For PCA decomposition (PCA - LDA)
from sklearn import \
    discriminant_analysis as da  # For LDA decomposition (PCA - LDA)
from sklearn.base import BaseEstimator
from sklearn.metrics import accuracy_score

# Used for naive bayes decoder
try:
    import statsmodels.api as sm
except ImportError:
    print(
        "\nWARNING: statsmodels is not installed. You will be unable to use "
        "the Naive Bayes Decoder")
    pass

try:
    from sklearnex import patch_sklearn

    # The names match scikit-learn estimators
    patch_sklearn(["PCA"])
except ImportError:
    print(
        "\nWARNING: sklearnex is not installed. You will be unable to use the"
        "PCA decoder acceleration")
    pass
# Import XGBoost if the package is installed
try:
    import xgboost as xgb  # For xgboost
except ImportError:
    print(
        "\nWARNING: Xgboost package is not installed. You will be unable to"
        "use the xgboost decoder")
    pass

# Import functions for Keras if Keras is installed
# Note that Keras has many more built-in functions that I have not imported
# because I have not used them but if you want to modify the decoders with
# other functions (e.g. regularization), import them here
try:
    import keras

    keras_v1 = int(keras.__version__[0]) <= 1
    from keras.models import Sequential
    from keras.layers import Dense, LSTM, SimpleRNN, GRU, Activation, Dropout
    from keras.utils import np_utils
except ImportError:
    print(
        "\nWARNING: Keras package is not installed. You will be unable to use"
        "all neural net decoders")
    pass


# %% DECODER FUNCTIONS


# %% WIENER FILTER

class WienerFilterRegression(object):
    """Class for the Wiener Filter Decoder

    There are no parameters to set.

    This simply leverages the scikit-learn linear regression.
    """

    def __init__(self):
        return

    def fit(self, X_flat_train, y_train):
        """Train Wiener Filter Decoder

        Parameters
        ----------
        X_flat_train: numpy 2d array of shape [n_samples,n_features]
            This is the neural data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 2d array of shape [n_samples, n_outputs]
            This is the outputs that are being predicted
        """
        # Initialize linear regression model
        self.model = linear_model.LinearRegression()

        # Train the model
        self.model.fit(X_flat_train, y_train)

    def predict(self, X_flat_test):
        """Predict outcomes using trained Wiener Cascade Decoder

        Parameters
        ----------
        X_flat_test: numpy 2d array of shape [n_samples,n_features]
            This is the neural data being used to predict outputs.

        Returns
        -------
        y_test_predicted: numpy 2d array of shape [n_samples,n_outputs]
            The predicted outputs
        """

        y_test_predicted = self.model.predict(X_flat_test)  # Make predictions
        return y_test_predicted


# %% WIENER CASCADE

class WienerCascadeRegression(object):
    """Class for the Wiener Cascade Decoder

    Parameters
    ----------
    degree: integer, optional, default 3
        The degree of the polynomial used for the static nonlinearity
    """

    def __init__(self, degree=3):
        self.degree = degree

    def fit(self, X_flat_train, y_train):

        """Train Wiener Cascade Decoder

        Parameters
        ----------
        X_flat_train: numpy 2d array of shape [n_samples,n_features]
            This is the neural data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 2d array of shape [n_samples, n_outputs]
            This is the outputs that are being predicted
        """

        num_outputs = y_train.shape[1]  # Number of outputs
        models = []  # Initialize list of models (there will be a separate
        # model for each output)
        for i in range(num_outputs):  # Loop through outputs
            # Fit linear portion of model
            regr = linear_model.LinearRegression()  # Call the linear portion
            # of the model "regr"
            regr.fit(X_flat_train, y_train[:, i])  # Fit linear
            y_train_predicted_linear = regr.predict(
                X_flat_train)  # Get outputs of linear portion of model
            # Fit nonlinear portion of model
            p = np.polyfit(y_train_predicted_linear, y_train[:, i],
                           self.degree)
            # Add model for this output (both linear and nonlinear parts)
            # to the list "models"
            models.append([regr, p])
        self.model = models

    def predict(self, X_flat_test):

        """Predict outcomes using trained Wiener Cascade Decoder

        Parameters
        ----------
        X_flat_test: numpy 2d array of shape [n_samples,n_features]
            This is the neural data being used to predict outputs.

        Returns
        -------
        y_test_predicted: numpy 2d array of shape [n_samples,n_outputs]
            The predicted outputs
        """

        num_outputs = len(
            self.model)  # Number of outputs being predicted. Recall from the
        # "fit" function that self.model is a list of models
        y_test_predicted = np.empty([X_flat_test.shape[0],
                                     num_outputs])  # Initialize matrix that
        # contains predicted outputs
        for i in range(num_outputs):  # Loop through outputs
            [regr, p] = self.model[
                i]  # Get the linear (regr) and nonlinear (p) portions of the
            # trained model
            # Predictions on test set
            y_test_predicted_linear = regr.predict(
                X_flat_test)  # Get predictions on the linear portion of
            # the model
            y_test_predicted[:, i] = np.polyval(p,
                                                y_test_predicted_linear)
            # Run the linear predictions through the nonlinearity to get
            # the final predictions
        return y_test_predicted


# %% KALMAN FILTER

class KalmanFilterRegression(object):
    """Class for the Kalman Filter Decoder

    Parameters
    -----------
    C - float, optional, default 1
    This parameter scales the noise matrix associated with the transition in
    kinematic states. It effectively allows changing the weight of the new
    neural evidence in the current update.

    Our implementation of the Kalman filter for neural decoding is based on
    that of Wu et al 2003 (https://papers.nips.cc/paper/2178-neural-decoding-of
    -cursor-motion-using-a-kalman-filter.pdf) with the exception of the
    addition of the parameter C. The original implementation has previously
    been coded in Matlab by Dan Morris
    (http://dmorris.net/projects/neural_decoding.html#code)
    """

    def __init__(self, C=1):
        self.C = C

    def fit(self, X_kf_train, y_train):
        """Train Kalman Filter Decoder

        Parameters
        ----------
        X_kf_train: numpy 2d array of shape [n_samples(i.e. timebins) ,
        n_neurons]
            This is the neural data in Kalman filter format.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 2d array of shape [n_samples(i.e. timebins), n_outputs]
            This is the outputs that are being predicted
        """

        # First we'll rename and reformat the variables to be in a more
        # standard kalman filter nomenclature (specifically that from Wu et
        # al, 2003):
        # xs are the state (here, the variable we're predicting, i.e. y_train)
        # zs are the observed variable (neural data here, i.e. X_kf_train)
        X = np.matrix(y_train.T)
        Z = np.matrix(X_kf_train.T)

        # number of time bins
        nt = X.shape[1]

        # Calculate the transition matrix (from x_t to x_t+1) using
        # least-squares, and compute its covariance
        # In our case, this is the transition from one kinematic state to
        # the next
        X2 = X[:, 1:]
        X1 = X[:, 0:nt - 1]
        A = X2 * X1.T * inv(X1 * X1.T)  # Transition matrix
        W = (X2 - A * X1) * (X2 - A * X1).T / (
                nt - 1) / self.C  # Covariance of transition matrix. Note we
        # divide by nt-1 since only nt-1 points were used in the computation
        # (that's the length of X1 and X2). We also introduce the extra
        # parameter C here.

        # Calculate the measurement matrix (from x_t to z_t) using
        # least-squares, and compute its covariance
        # In our case, this is the transformation from kinematics to spikes
        H = Z * X.T * (inv(X * X.T))  # Measurement matrix
        Q = ((Z - H * X) * (
            (Z - H * X).T)) / nt  # Covariance of measurement matrix
        params = [A, W, H, Q]
        self.model = params

    def predict(self, X_kf_test, y_test):
        """Predict outcomes using trained Kalman Filter Decoder

        Parameters
        ----------
        X_kf_test: numpy 2d array of shape [n_samples(i.e. timebins) ,
        n_neurons]
            This is the neural data in Kalman filter format.

        y_test: numpy 2d array of shape [n_samples(i.e. timebins),n_outputs]
            The actual outputs
            This parameter is necesary for the Kalman filter (unlike other
            decoders)
            because the first value is nececessary for initialization

        Returns
        -------
        y_test_predicted: numpy 2d array of shape [n_samples(i.e. timebins),
        n_outputs]
            The predicted outputs
        """

        # Extract parameters
        A, W, H, Q = self.model

        # First we'll rename and reformat the variables to be in a more
        # standard kalman filter nomenclature (specifically that from Wu et
        # al):
        # xs are the state (here, the variable we're predicting, i.e. y_train)
        # zs are the observed variable (neural data here, i.e. X_kf_train)
        X = np.matrix(y_test.T)
        Z = np.matrix(X_kf_test.T)

        # Initializations
        num_states = X.shape[0]  # Dimensionality of the state
        states = np.empty(
            X.shape)  # Keep track of states over time (states is what will
        # be returned as y_test_predicted)
        P_m = np.matrix(np.zeros([num_states, num_states]))
        P = np.matrix(np.zeros([num_states, num_states]))
        state = X[:, 0]  # Initial state
        states[:, 0] = np.copy(np.squeeze(state))

        # Get predicted state for every time bin
        for t in range(X.shape[1] - 1):
            # Do first part of state update - based on transition matrix
            P_m = A * P * A.T + W
            state_m = A * state

            # Do second part of state update - based on measurement matrix
            K = P_m * H.T * inv(H * P_m * H.T + Q)  # Calculate Kalman gain
            P = (np.matrix(np.eye(num_states)) - K * H) * P_m
            state = state_m + K * (Z[:, t + 1] - H * state_m)
            states[:, t + 1] = np.squeeze(
                state)  # Record state at the timestep
        y_test_predicted = states.T
        return y_test_predicted


# %% DENSE (FULLY-CONNECTED) NEURAL NETWORK

class DenseNNRegression(object):
    """Class for the dense (fully-connected) neural network decoder

    Parameters
    ----------

    units: integer or vector of integers, optional, default 400
        This is the number of hidden units in each layer
        If you want a single layer, input an integer (e.g. units=400 will give
        you a single hidden layer with 400 units) If you want multiple layers,
        input a vector (e.g. units=[400,200]) will give you 2 hidden layers
        with 400 and 200 units, repsectively.
        The vector can either be a list or an array

    dropout: decimal, optional, default 0
        Proportion of units that get dropped out

    num_epochs: integer, optional, default 10
        Number of epochs used for training

    verbose: binary, optional, default=0
        Whether to show progress of the fit after each epoch
    """

    def __init__(self, units=400, dropout=0, num_epochs=10, verbose=0):
        self.dropout = dropout
        self.num_epochs = num_epochs
        self.verbose = verbose

        # If "units" is an integer, put it in the form of a vector
        try:  # Check if it's a vector
            units[0]
        except IndexError:
            # If it's not a vector, create a vector of the number of
            # units for each layer
            units = [units]
        self.units = units

        # Determine the number of hidden layers (based on "units" that the
        # user entered)
        self.num_layers = len(units)

    def fit(self, X_flat_train, y_train):

        """Train DenseNN Decoder

        Parameters
        ----------
        X_flat_train: numpy 2d array of shape [n_samples,n_features]
            This is the neural data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 2d array of shape [n_samples, n_outputs]
            This is the outputs that are being predicted
        """

        model = Sequential()  # Declare model
        # Add first hidden layer
        model.add(Dense(self.units[0],
                        input_dim=X_flat_train.shape[1]))  # Add dense layer
        model.add(Activation('relu'))  # Add nonlinear (tanh) activation
        # if self.dropout!=0:
        if self.dropout != 0:
            # Dropout some units if proportion of dropout != 0
            model.add(Dropout(self.dropout))

        # Add any additional hidden layers (beyond the 1st)
        for layer in range(
                self.num_layers - 1):  # Loop through additional layers
            model.add(Dense(self.units[layer + 1]))  # Add dense layer
            model.add(Activation('relu'))  # Add nonlinear (tanh) activation
            if self.dropout != 0:
                # Dropout some units if proportion of dropout != 0
                model.add(Dropout(self.dropout))

        # Add dense connections to all outputs
        model.add(Dense(
            y_train.shape[1]))  # Add final dense layer (connected to outputs)

        # Fit model (and set fitting parameters)
        model.compile(loss='mse', optimizer='adam',
                      metrics=['accuracy'])  # Set loss function and optimizer
        if keras_v1:
            model.fit(X_flat_train, y_train, nb_epoch=self.num_epochs,
                      verbose=self.verbose)  # Fit the model
        else:
            model.fit(X_flat_train, y_train, epochs=self.num_epochs,
                      verbose=self.verbose)  # Fit the model
        self.model = model

    def predict(self, X_flat_test):

        """Predict outcomes using trained DenseNN Decoder

        Parameters
        ----------
        X_flat_test: numpy 2d array of shape [n_samples,n_features]
            This is the neural data being used to predict outputs.

        Returns
        -------
        y_test_predicted: numpy 2d array of shape [n_samples,n_outputs]
            The predicted outputs
        """

        y_test_predicted = self.model.predict(X_flat_test)  # Make predictions
        return y_test_predicted


# %% SIMPLE RECURRENT NEURAL NETWORK

class SimpleRNNRegression(object):
    """Class for the simple recurrent neural network decoder

    Parameters
    ----------
    units: integer, optional, default 400
        Number of hidden units in each layer

    dropout: decimal, optional, default 0
        Proportion of units that get dropped out

    num_epochs: integer, optional, default 10
        Number of epochs used for training

    verbose: binary, optional, default=0
        Whether to show progress of the fit after each epoch
    """

    def __init__(self, units=400, dropout=0, num_epochs=10, verbose=0):
        self.units = units
        self.dropout = dropout
        self.num_epochs = num_epochs
        self.verbose = verbose

    def fit(self, X_train, y_train):

        """Train SimpleRNN Decoder

        Parameters
        ----------
        X_train: numpy 3d array of shape [n_samples,n_time_bins,n_neurons]
            This is the neural data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 2d array of shape [n_samples, n_outputs]
            This is the outputs that are being predicted
        """

        model = Sequential()  # Declare model
        # Add recurrent layer
        if keras_v1:
            model.add(SimpleRNN(self.units, input_shape=(
                X_train.shape[1], X_train.shape[2]), dropout_W=self.dropout,
                                dropout_U=self.dropout,
                                activation='relu'))  # Within recurrent
            # layer, include dropout
        else:
            model.add(SimpleRNN(self.units, input_shape=(
                X_train.shape[1], X_train.shape[2]), dropout=self.dropout,
                                recurrent_dropout=self.dropout,
                                activation='relu'))  # Within recurrent
            # layer, include dropout
        if self.dropout != 0:
            # Dropout some units (recurrent layer output units)
            model.add(Dropout(self.dropout))

        # Add dense connections to output layer
        model.add(Dense(y_train.shape[1]))

        # Fit model (and set fitting parameters)
        model.compile(loss='mse', optimizer='rmsprop',
                      metrics=['accuracy'])  # Set loss function and optimizer
        if keras_v1:
            model.fit(X_train, y_train, nb_epoch=self.num_epochs,
                      verbose=self.verbose)  # Fit the model
        else:
            model.fit(X_train, y_train, epochs=self.num_epochs,
                      verbose=self.verbose)  # Fit the model
        self.model = model

    def predict(self, X_test):

        """Predict outcomes using trained SimpleRNN Decoder

        Parameters
        ----------
        X_test: numpy 3d array of shape [n_samples,n_time_bins,n_neurons]
            This is the neural data being used to predict outputs.

        Returns
        -------
        y_test_predicted: numpy 2d array of shape [n_samples,n_outputs]
            The predicted outputs
        """

        y_test_predicted = self.model.predict(X_test)  # Make predictions
        return y_test_predicted


# %% GATED RECURRENT UNIT (GRU) DECODER

class GRURegression(object):
    """Class for the gated recurrent unit (GRU) decoder

    Parameters
    ----------
    units: integer, optional, default 400
        Number of hidden units in each layer

    dropout: decimal, optional, default 0
        Proportion of units that get dropped out

    num_epochs: integer, optional, default 10
        Number of epochs used for training

    verbose: binary, optional, default=0
        Whether to show progress of the fit after each epoch
    """

    def __init__(self, units=400, dropout=0, num_epochs=10, verbose=0):
        self.units = units
        self.dropout = dropout
        self.num_epochs = num_epochs
        self.verbose = verbose

    def fit(self, X_train, y_train):

        """Train GRU Decoder

        Parameters
        ----------
        X_train: numpy 3d array of shape [n_samples,n_time_bins,n_neurons]
            This is the neural data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 2d array of shape [n_samples, n_outputs]
            This is the outputs that are being predicted
        """

        model = Sequential()  # Declare model
        # Add recurrent layer
        if keras_v1:
            model.add(GRU(self.units,
                          input_shape=(X_train.shape[1], X_train.shape[2]),
                          dropout_W=self.dropout,
                          dropout_U=self.dropout))  # Within recurrent layer,
            # include dropout
        else:
            model.add(GRU(self.units,
                          input_shape=(X_train.shape[1], X_train.shape[2]),
                          dropout=self.dropout,
                          recurrent_dropout=self.dropout))
        if self.dropout != 0:
            # Dropout some units (recurrent layer output units)
            model.add(Dropout(self.dropout))

        # Add dense connections to output layer
        model.add(Dense(y_train.shape[1]))

        # Fit model (and set fitting parameters)
        model.compile(loss='mse', optimizer='rmsprop',
                      metrics=['accuracy'])  # Set loss function and optimizer
        if keras_v1:
            model.fit(X_train, y_train, nb_epoch=self.num_epochs,
                      verbose=self.verbose)  # Fit the model
        else:
            model.fit(X_train, y_train, epochs=self.num_epochs,
                      verbose=self.verbose)  # Fit the model
        self.model = model

    def predict(self, X_test):

        """Predict outcomes using trained GRU Decoder

        Parameters
        ----------
        X_test: numpy 3d array of shape [n_samples,n_time_bins,n_neurons]
            This is the neural data being used to predict outputs.

        Returns
        -------
        y_test_predicted: numpy 2d array of shape [n_samples,n_outputs]
            The predicted outputs
        """

        y_test_predicted = self.model.predict(X_test)  # Make predictions
        return y_test_predicted


# %% LONG SHORT TERM MEMORY (LSTM) DECODER

class LSTMRegression(object):
    """Class for the gated recurrent unit (GRU) decoder

    Parameters
    ----------
    units: integer, optional, default 400
        Number of hidden units in each layer

    dropout: decimal, optional, default 0
        Proportion of units that get dropped out

    num_epochs: integer, optional, default 10
        Number of epochs used for training

    verbose: binary, optional, default=0
        Whether to show progress of the fit after each epoch
    """

    def __init__(self, units=400, dropout=0, num_epochs=10, verbose=0):
        self.units = units
        self.dropout = dropout
        self.num_epochs = num_epochs
        self.verbose = verbose

    def fit(self, X_train, y_train):

        """Train LSTM Decoder

        Parameters
        ----------
        X_train: numpy 3d array of shape [n_samples,n_time_bins,n_neurons]
            This is the neural data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 2d array of shape [n_samples, n_outputs]
            This is the outputs that are being predicted
        """

        model = Sequential()  # Declare model
        # Add recurrent layer
        if keras_v1:
            model.add(LSTM(self.units,
                           input_shape=(X_train.shape[1], X_train.shape[2]),
                           dropout_W=self.dropout,
                           dropout_U=self.dropout))  # Within recurrent layer,
            # include dropout
        else:
            model.add(LSTM(self.units,
                           input_shape=(X_train.shape[1], X_train.shape[2]),
                           dropout=self.dropout,
                           recurrent_dropout=self.dropout))  # Within recurrent
            # layer, include dropout
        if self.dropout != 0:
            # Dropout some units (recurrent layer output units)
            model.add(Dropout(self.dropout))

        # Add dense connections to output layer
        model.add(Dense(y_train.shape[1]))

        # Fit model (and set fitting parameters)
        model.compile(loss='mse', optimizer='rmsprop',
                      metrics=['accuracy'])  # Set loss function and optimizer
        if keras_v1:
            model.fit(X_train, y_train, nb_epoch=self.num_epochs,
                      verbose=self.verbose)  # Fit the model
        else:
            model.fit(X_train, y_train, epochs=self.num_epochs,
                      verbose=self.verbose)  # Fit the model
        self.model = model

    def predict(self, X_test):

        """Predict outcomes using trained LSTM Decoder

        Parameters
        ----------
        X_test: numpy 3d array of shape [n_samples,n_time_bins,n_neurons]
            This is the neural data being used to predict outputs.

        Returns
        -------
        y_test_predicted: numpy 2d array of shape [n_samples,n_outputs]
            The predicted outputs
        """

        y_test_predicted = self.model.predict(X_test)  # Make predictions
        return y_test_predicted


# %% EXTREME GRADIENT BOOSTING (XGBOOST)

class XGBoostRegression(object):
    """Class for the XGBoost Decoder

    Parameters
    ----------
    max_depth: integer, optional, default=3
        the maximum depth of the trees

    num_round: integer, optional, default=300
        the number of trees that are fit

    eta: float, optional, default=0.3
        the learning rate

    gpu: integer, optional, default=-1
        if the gpu version of xgboost is installed, this can be used to select
        which gpu to use
        for negative values (default), the gpu is not used
    """

    def __init__(self, max_depth=3, num_round=300, eta=0.3, gpu=-1):
        self.max_depth = max_depth
        self.num_round = num_round
        self.eta = eta
        self.gpu = gpu

    def fit(self, X_flat_train, y_train):

        """Train XGBoost Decoder

        Parameters
        ----------
        X_flat_train: numpy 2d array of shape [n_samples,n_features]
            This is the neural data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 2d array of shape [n_samples, n_outputs]
            This is the outputs that are being predicted
        """

        num_outputs = y_train.shape[1]  # Number of outputs

        # Set parameters for XGBoost
        param = {'objective': "reg:linear",  # for linear output
                 'eval_metric': "logloss",  # loglikelihood loss
                 'max_depth': self.max_depth,
                 # this is the only parameter we have set, it's one of the way
                 # or regularizing
                 'eta': self.eta,
                 'seed': 2925,  # for reproducibility
                 'silent': 1}
        if self.gpu < 0:
            param['nthread'] = -1  # with -1 it will use all available threads
        else:
            param['gpu_id'] = self.gpu
            param['updater'] = 'grow_gpu'

        models = []  # Initialize list of models (there will be a separate
        # model for each output)
        for y_idx in range(num_outputs):  # Loop through outputs
            dtrain = xgb.DMatrix(X_flat_train, label=y_train[:, y_idx])
            # Put in correct format for XGB
            bst = xgb.train(param, dtrain, self.num_round)  # Train model
            models.append(bst)  # Add fit model to list of models

        self.model = models

    def predict(self, X_flat_test):

        """Predict outcomes using trained XGBoost Decoder

        Parameters
        ----------
        X_flat_test: numpy 2d array of shape [n_samples,n_features]
            This is the neural data being used to predict outputs.

        Returns
        -------
        y_test_predicted: numpy 2d array of shape [n_samples,n_outputs]
            The predicted outputs
        """

        dtest = xgb.DMatrix(X_flat_test)  # Put in XGB format
        num_outputs = len(self.model)  # Number of outputs
        y_test_predicted = np.empty([X_flat_test.shape[0],
                                     num_outputs])  # Initialize matrix of
        # predicted outputs
        for y_idx in range(num_outputs):  # Loop through outputs
            bst = self.model[y_idx]  # Get fit model for this output
            y_test_predicted[:, y_idx] = bst.predict(dtest)  # Make prediction
        return y_test_predicted


# %% SUPPORT VECTOR REGRESSION

class SVRegression(object):
    """Class for the Support Vector Regression (SVR) Decoder
    This simply leverages the scikit-learn SVR

    Parameters
    ----------
    C: float, default=3.0
        Penalty parameter of the error term

    max_iter: integer, default=-1
        the maximum number of iteraations to run (to save time)
        max_iter=-1 means no limit
        Typically in the 1000s takes a short amount of time on a laptop
    """

    def __init__(self, max_iter=-1, C=3.0):
        self.max_iter = max_iter
        self.C = C
        return

    def fit(self, X_flat_train, y_train):

        """Train SVR Decoder

        Parameters
        ----------
        X_flat_train: numpy 2d array of shape [n_samples,n_features]
            This is the neural data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 2d array of shape [n_samples, n_outputs]
            This is the outputs that are being predicted
        """

        num_outputs = y_train.shape[1]  # Number of outputs
        models = []  # Initialize list of models (there will be a separate
        # model for each output)
        for y_idx in range(num_outputs):  # Loop through outputs
            model = SVR(C=self.C,
                        max_iter=self.max_iter)  # Initialize SVR model
            model.fit(X_flat_train, y_train[:, y_idx])  # Train the model
            models.append(model)  # Add fit model to list of models
        self.model = models

    def predict(self, X_flat_test):

        """Predict outcomes using trained SVR Decoder

        Parameters
        ----------
        X_flat_test: numpy 2d array of shape [n_samples,n_features]
            This is the neural data being used to predict outputs.

        Returns
        -------
        y_test_predicted: numpy 2d array of shape [n_samples,n_outputs]
            The predicted outputs
        """

        num_outputs = len(self.model)  # Number of outputs
        y_test_predicted = np.empty([X_flat_test.shape[0],
                                     num_outputs])  # Initialize matrix of
        # predicted outputs
        for y_idx in range(num_outputs):  # Loop through outputs
            model = self.model[y_idx]  # Get fit model for that output
            y_test_predicted[:, y_idx] = model.predict(
                X_flat_test)  # Make predictions
        return y_test_predicted


# GLM helper function for the NaiveBayesDecoder
def glm_run(Xr, Yr, X_range):
    X2 = sm.add_constant(Xr)

    poiss_model = sm.GLM(Yr, X2, family=sm.families.Poisson())
    try:
        glm_results = poiss_model.fit()
        Y_range = glm_results.predict(sm.add_constant(X_range))
    except np.linalg.LinAlgError:
        print("\nWARNING: LinAlgError")
        Y_range = np.mean(Yr) * np.ones([X_range.shape[0], 1])

    return Y_range


class NaiveBayesRegression(object):
    """Class for the Naive Bayes Decoder

    Parameters
    ----------
    encoding_model: string, default='quadratic'
        what encoding model is used

    res:int, default=100
        resolution of predicted values
        This is the number of bins to divide the outputs into (going from
        minimum to maximum) larger values will make decoding slower
    """

    def __init__(self, encoding_model='quadratic', res=100):
        self.encoding_model = encoding_model
        self.res = res
        return

    def fit(self, X_b_train, y_train):

        """Train Naive Bayes Decoder

        Parameters
        ----------
        X_b_train: numpy 2d array of shape [n_samples,n_neurons]
            This is the neural training data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 2d array of shape [n_samples, n_outputs]
            This is the outputs that are being predicted (training data)
        """

        # %% FIT TUNING CURVE
        # First, get the output values (x/y position or velocity) that we will
        # be creating tuning curves over
        # Create the range for x and y (position/velocity) values
        input_x_range = np.arange(np.min(y_train[:, 0]),
                                  np.max(y_train[:, 0]) + .01, np.round(
                (np.max(y_train[:, 0]) - np.min(y_train[:, 0])) / self.res))
        input_y_range = np.arange(np.min(y_train[:, 1]),
                                  np.max(y_train[:, 1]) + .01, np.round(
                (np.max(y_train[:, 1]) - np.min(y_train[:, 1])) / self.res))
        # Get all combinations of x/y values
        input_mat = np.meshgrid(input_x_range, input_y_range)
        # Format so that all combinations of x/y values are in 2 columns (first
        # column x, second column y). This is called "input_xy"
        xs = np.reshape(input_mat[0],
                        [input_x_range.shape[0] * input_y_range.shape[0], 1])
        ys = np.reshape(input_mat[1],
                        [input_x_range.shape[0] * input_y_range.shape[0], 1])
        input_xy = np.concatenate((xs, ys), axis=1)

        # If quadratic model:
        #   -make covariates have squared components and mixture of x and y
        #   -do same thing for "input_xy", which are the values for creating
        #   the tuning curves
        if self.encoding_model == 'quadratic':
            input_xy_modified = np.empty([input_xy.shape[0], 5])
            input_xy_modified[:, 0] = input_xy[:, 0] ** 2
            input_xy_modified[:, 1] = input_xy[:, 0]
            input_xy_modified[:, 2] = input_xy[:, 1] ** 2
            input_xy_modified[:, 3] = input_xy[:, 1]
            input_xy_modified[:, 4] = input_xy[:, 0] * input_xy[:, 1]
            y_train_modified = np.empty([y_train.shape[0], 5])
            y_train_modified[:, 0] = y_train[:, 0] ** 2
            y_train_modified[:, 1] = y_train[:, 0]
            y_train_modified[:, 2] = y_train[:, 1] ** 2
            y_train_modified[:, 3] = y_train[:, 1]
            y_train_modified[:, 4] = y_train[:, 0] * y_train[:, 1]

        # Create tuning curves

        num_nrns = X_b_train.shape[
            1]  # Number of neurons to fit tuning curves for
        tuning_all = np.zeros([num_nrns, input_xy.shape[
            0]])  # Matrix that stores tuning curves for all neurons

        # Loop through neurons and fit tuning curves
        for j in range(num_nrns):  # Neuron number

            if self.encoding_model == 'linear':
                tuning = glm_run(y_train, X_b_train[:, j:j + 1], input_xy)
            if self.encoding_model == 'quadratic':
                tuning = glm_run(y_train_modified, X_b_train[:, j:j + 1],
                                 input_xy_modified)
            # Enter tuning curves into matrix
            tuning_all[j, :] = np.squeeze(tuning)

        # Save tuning curves to be used in "predict" function
        self.tuning_all = tuning_all
        self.input_xy = input_xy

        # Get information about the probability of being in one state
        # (position/velocity) based on the previous state. Here we're
        # calculating the standard deviation of the change in state
        # (velocity/acceleration) in the training set
        n = y_train.shape[0]
        dx = np.zeros([n - 1, 1])
        for i in range(n - 1):
            # Change in state across time steps
            dx[i] = np.sqrt((y_train[i + 1, 0] - y_train[i, 0]) ** 2 + (
                    y_train[i + 1, 1] - y_train[i, 1]) ** 2)
        std = np.sqrt(np.mean(
            dx ** 2))  # dx is only positive. this gets approximate stdev of
        # distribution (if it was positive and negative)
        self.std = std  # Save for use in "predict" function

        # Get probability of being in each state - we are not using this since
        # it did not help decoding performance
        # n_x=np.empty([input_xy.shape[0]])
        # for i in range(n):
        #     loc_idx=np.argmin(cdist(y_train[0:1,:],input_xy))
        #     n_x[loc_idx]=n_x[loc_idx]+1
        # p_x=n_x/n
        # self.p_x=p_x

    def predict(self, X_b_test, y_test):

        """Predict outcomes using trained tuning curves

        Parameters
        ----------
        X_b_test: numpy 2d array of shape [n_samples,n_features]
            This is the neural data being used to predict outputs.

        y_test: numpy 2d array of shape [n_samples,n_outputs]
            The actual outputs
            This parameter is necesary for the NaiveBayesDecoder  (unlike most
            other decoders) because the first value is nececessary for
            initialization

        Returns
        -------
        y_test_predicted: numpy 2d array of shape [n_samples,n_outputs]
            The predicted outputs
        """

        # Get values saved in "fit" function
        tuning_all = self.tuning_all
        input_xy = self.input_xy
        std = self.std

        # Get probability of going from one state to the next
        dists = squareform(pdist(input_xy,
                                 'euclidean'))
        # Distance between all states in "input_xy" Probability of going from
        # one state to the next, based on the above calculated distances
        # The probability is calculated based on the distances coming from a
        # Gaussian with standard deviation of std
        prob_dists = norm.pdf(dists, 0, std)

        # Initializations
        loc_idx = np.argmin(
            cdist(y_test[0:1, :], input_xy))  # The index of the first location
        num_nrns = tuning_all.shape[0]  # Number of neurons
        y_test_predicted = np.empty(
            [X_b_test.shape[0], 2])  # Initialize matrix of predicted outputs
        num_ts = X_b_test.shape[0]  # Number of time steps we are predicting

        # Loop across time and decode
        for t in range(num_ts):
            rs = X_b_test[t, :]
            # Number of spikes at this time point (in the interval
            # we've specified including bins_before and bins_after)

            probs_total = np.ones([tuning_all[0, :].shape[
                                       0]])
            # Vector that stores the probabilities of being in any state based
            # on the neural activity (does not include probabilities of going
            # from one state to the next)
            for j in range(num_nrns):  # Loop across neurons
                lam = np.copy(tuning_all[j, :])
                # Expected spike counts given the tuning curve
                r = rs[j]  # Actual spike count
                probs = np.exp(-lam) * lam ** r / math.factorial(r)
                # Probability of the given neuron's spike count given tuning
                # curve (assuming poisson distribution)
                probs_total = np.copy(probs_total * probs)
                # Update the probability across neurons (probabilities are
                # multiplied across neurons due to the independence assumption)
            prob_dists_vec = np.copy(prob_dists[loc_idx, :])
            # Probability of going to all states from the previous state
            probs_final = probs_total * prob_dists_vec
            # Get final probability (multiply probabilities based on spike
            # count and previous state)
            # probs_final=probs_total*prob_dists_vec*self.p_x
            # #Get final probability when including p(x), i.e. prior about
            # being in states, which we're not using
            loc_idx = np.argmax(probs_final)
            # Get the index of the current state (that w/ the highest
            # probability)
            y_test_predicted[t, :] = input_xy[loc_idx, :]
            # The current predicted output

        return y_test_predicted  # Return predictions


# %% ALIASES for Regression

WienerFilterDecoder = WienerFilterRegression
WienerCascadeDecoder = WienerCascadeRegression
KalmanFilterDecoder = KalmanFilterRegression
DenseNNDecoder = DenseNNRegression
SimpleRNNDecoder = SimpleRNNRegression
GRUDecoder = GRURegression
LSTMDecoder = LSTMRegression
XGBoostDecoder = XGBoostRegression
SVRDecoder = SVRegression
NaiveBayesDecoder = NaiveBayesRegression


# %% CLASSIFICATION


class WienerFilterClassification(object):
    """Class for the Wiener Filter Decoder

    There are no parameters to set.

    This simply leverages the scikit-learn logistic regression.
    """

    def __init__(self, C=1):
        self.C = C
        return

    def fit(self, X_flat_train, y_train):
        """Train Wiener Filter Decoder

        Parameters
        ----------
        X_flat_train: numpy 2d array of shape [n_samples,n_features]
            This is the neural data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 2d array of shape [n_samples, n_outputs]
            This is the outputs that are being predicted
        """

        # if self.C>0:
        self.model = linear_model.LogisticRegression(C=self.C,
                                                     multi_class='auto')
        # Initialize linear regression model
        # else:
        # self.model=linear_model.LogisticRegression(penalty='none',
        # solver='newton-cg') #Initialize linear regression model
        self.model.fit(X_flat_train, y_train)  # Train the model

    def predict(self, X_flat_test):
        """Predict outcomes using trained Wiener Cascade Decoder

        Parameters
        ----------
        X_flat_test: numpy 2d array of shape [n_samples,n_features]
            This is the neural data being used to predict outputs.

        Returns
        -------
        y_test_predicted: numpy 2d array of shape [n_samples,n_outputs]
            The predicted outputs
        """

        y_test_predicted = self.model.predict(X_flat_test)  # Make predictions
        return y_test_predicted


# %% SUPPORT VECTOR REGRESSION

class SVClassification(object):
    """Class for the Support Vector Classification Decoder
    This simply leverages the scikit-learn SVM

    Parameters
    ----------
    C: float, default=3.0
        Penalty parameter of the error term

    max_iter: integer, default=-1
        the maximum number of iteraations to run (to save time)
        max_iter=-1 means no limit
        Typically in the 1000s takes a short amount of time on a laptop
    """

    def __init__(self, max_iter=-1, C=3.0):
        self.max_iter = max_iter
        self.C = C
        return

    def fit(self, X_flat_train, y_train):
        """Train SVR Decoder

        Parameters
        ----------
        X_flat_train: numpy 2d array of shape [n_samples,n_features]
            This is the neural data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 2d array of shape [n_samples, n_outputs]
            This is the outputs that are being predicted
        """

        model = SVC(C=self.C, max_iter=self.max_iter)  # Initialize model
        model.fit(X_flat_train, y_train)  # Train the model
        self.model = model

    def predict(self, X_flat_test):
        """Predict outcomes using trained SV Decoder

        Parameters
        ----------
        X_flat_test: numpy 2d array of shape [n_samples,n_features]
            This is the neural data being used to predict outputs.

        Returns
        -------
        y_test_predicted: numpy 2d array of shape [n_samples,n_outputs]
            The predicted outputs
        """

        model = self.model  # Get fit model for that output
        y_test_predicted = model.predict(X_flat_test)  # Make predictions
        return y_test_predicted


# %% DENSE (FULLY-CONNECTED) NEURAL NETWORK

class DenseNNClassification(object):
    """Class for the dense (fully-connected) neural network decoder

    Parameters
    ----------

    units: integer or vector of integers, optional, default 400
        This is the number of hidden units in each layer
        If you want a single layer, input an integer (e.g. units=400 will give
        you a single hidden layer with 400 units). If you want multiple layers,
        input a vector (e.g. units=[400,200]) will give you 2 hidden layers
        with 400 and 200 units, repsectively. The vector can either be a list
        or an array

    dropout: decimal, optional, default 0
        Proportion of units that get dropped out

    num_epochs: integer, optional, default 10
        Number of epochs used for training

    verbose: binary, optional, default=0
        Whether to show progress of the fit after each epoch
    """

    def __init__(self, units=400, dropout=0, num_epochs=10, verbose=0):
        self.dropout = dropout
        self.num_epochs = num_epochs
        self.verbose = verbose

        # If "units" is an integer, put it in the form of a vector
        try:  # Check if it's a vector
            units[0]
        except IndexError:
            # If it's not a vector, create a vector of the number of
            # units for each layer
            units = [units]
        self.units = units

        # Determine the number of hidden layers (based on "units" that the
        # user entered)
        self.num_layers = len(units)

    def fit(self, X_flat_train, y_train):

        """Train DenseNN Decoder

        Parameters
        ----------
        X_flat_train: numpy 2d array of shape [n_samples,n_features]
            This is the neural data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 2d array of shape [n_samples, n_outputs]
            This is the outputs that are being predicted
        """

        # Use one-hot coding for y
        if y_train.ndim == 1:
            y_train = np_utils.to_categorical(y_train.astype(int))
        elif y_train.shape[1] == 1:
            y_train = np_utils.to_categorical(y_train.astype(int))

        model = Sequential()  # Declare model
        # Add first hidden layer
        model.add(Dense(self.units[0],
                        input_dim=X_flat_train.shape[1]))  # Add dense layer
        model.add(Activation('relu'))  # Add nonlinear (tanh) activation
        # if self.dropout!=0:
        if self.dropout != 0:
            # Dropout some units if proportion of dropout != 0
            model.add(Dropout(self.dropout))

        # Add any additional hidden layers (beyond the 1st)
        for layer in range(
                self.num_layers - 1):  # Loop through additional layers
            model.add(Dense(self.units[layer + 1]))  # Add dense layer
            # Add nonlinear (tanh) activation - can also make
            model.add(Activation('tanh'))
            # relu
            if self.dropout != 0:
                # Dropout some units if proportion of dropout != 0
                model.add(Dropout(self.dropout))

        # Add dense connections to all outputs
        model.add(Dense(
            y_train.shape[1]))  # Add final dense layer (connected to outputs)
        model.add(Activation('softplus'))

        # Fit model (and set fitting parameters)
        model.compile(loss='categorical_crossentropy', optimizer='adam',
                      metrics=['accuracy'])  # Set loss function and optimizer
        if keras_v1:
            model.fit(X_flat_train, y_train, nb_epoch=self.num_epochs,
                      verbose=self.verbose)  # Fit the model
        else:
            model.fit(X_flat_train, y_train, epochs=self.num_epochs,
                      verbose=self.verbose)  # Fit the model
            self.model = model

    def predict(self, X_flat_test):

        """Predict outcomes using trained DenseNN Decoder

        Parameters
        ----------
        X_flat_test: numpy 2d array of shape [n_samples,n_features]
            This is the neural data being used to predict outputs.

        Returns
        -------
        y_test_predicted: numpy 2d array of shape [n_samples,n_outputs]
            The predicted outputs
        """

        y_test_predicted_raw = self.model.predict(
            X_flat_test)  # Make predictions

        y_test_predicted = np.argmax(y_test_predicted_raw, axis=1)

        return y_test_predicted


# %% SIMPLE RNN DECODER

class SimpleRNNClassification(object):
    """Class for the RNN decoder

    Parameters
    ----------
    units: integer, optional, default 400
        Number of hidden units in each layer

    dropout: decimal, optional, default 0
        Proportion of units that get dropped out

    num_epochs: integer, optional, default 10
        Number of epochs used for training

    verbose: binary, optional, default=0
        Whether to show progress of the fit after each epoch
    """

    def __init__(self, units=400, dropout=0, num_epochs=10, verbose=0):
        self.units = units
        self.dropout = dropout
        self.num_epochs = num_epochs
        self.verbose = verbose

    def fit(self, X_train, y_train):

        """Train GRU Decoder

        Parameters
        ----------
        X_train: numpy 3d array of shape [n_samples,n_time_bins,n_neurons]
            This is the neural data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 2d array of shape [n_samples, n_outputs]
            This is the outputs that are being predicted
        """

        # Use one-hot coding for y
        if y_train.ndim == 1:
            y_train = np_utils.to_categorical(y_train.astype(int))
        elif y_train.shape[1] == 1:
            y_train = np_utils.to_categorical(y_train.astype(int))

        model = Sequential()  # Declare model
        # Add recurrent layer

        # %% MAKE RELU ACTIVATION BELOW LIKE IN REGRESSION?????
        if keras_v1:
            model.add(SimpleRNN(self.units, input_shape=(
                X_train.shape[1], X_train.shape[2]), dropout_W=self.dropout,
                                dropout_U=self.dropout))
            # Within recurrent layer, include dropout
        else:
            model.add(SimpleRNN(self.units, input_shape=(
                X_train.shape[1], X_train.shape[2]), dropout=self.dropout,
                                recurrent_dropout=self.dropout))
            # Within recurrent layer, include dropout
        if self.dropout != 0:
            # Dropout some units (recurrent layer output units)
            model.add(Dropout(self.dropout))

        # Add dense connections to output layer
        model.add(Dense(y_train.shape[1]))
        model.add(Activation('softplus'))

        # Fit model (and set fitting parameters)
        model.compile(loss='categorical_crossentropy', optimizer='rmsprop',
                      metrics=['accuracy'])  # Set loss function and optimizer
        if keras_v1:
            model.fit(X_train, y_train, nb_epoch=self.num_epochs,
                      verbose=self.verbose)  # Fit the model
        else:
            model.fit(X_train, y_train, epochs=self.num_epochs,
                      verbose=self.verbose)  # Fit the model
        self.model = model

    def predict(self, X_test):

        """Predict outcomes using trained LSTM Decoder

        Parameters
        ----------
        X_test: numpy 3d array of shape [n_samples,n_time_bins,n_neurons]
            This is the neural data being used to predict outputs.

        Returns
        -------
        y_test_predicted: numpy 2d array of shape [n_samples,n_outputs]
            The predicted outputs
        """

        y_test_predicted_raw = self.model.predict(X_test)  # Make predictions
        y_test_predicted = np.argmax(y_test_predicted_raw, axis=1)

        return y_test_predicted


# %% GATED RECURRENT UNIT (GRU) DECODER

class GRUClassification(object):
    """Class for the gated recurrent unit (GRU) decoder

    Parameters
    ----------
    units: integer, optional, default 400
        Number of hidden units in each layer

    dropout: decimal, optional, default 0
        Proportion of units that get dropped out

    num_epochs: integer, optional, default 10
        Number of epochs used for training

    verbose: binary, optional, default=0
        Whether to show progress of the fit after each epoch
    """

    def __init__(self, units=400, dropout=0, num_epochs=10, verbose=0):
        self.units = units
        self.dropout = dropout
        self.num_epochs = num_epochs
        self.verbose = verbose

    def fit(self, X_train, y_train):

        """Train GRU Decoder

        Parameters
        ----------
        X_train: numpy 3d array of shape [n_samples,n_time_bins,n_neurons]
            This is the neural data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 2d array of shape [n_samples, n_outputs]
            This is the outputs that are being predicted
        """

        # Use one-hot coding for y
        if y_train.ndim == 1:
            y_train = np_utils.to_categorical(y_train.astype(int))
        elif y_train.shape[1] == 1:
            y_train = np_utils.to_categorical(y_train.astype(int))

        model = Sequential()  # Declare model
        # Add recurrent layer
        if keras_v1:
            model.add(GRU(self.units,
                          input_shape=(X_train.shape[1], X_train.shape[2]),
                          dropout_W=self.dropout,
                          dropout_U=self.dropout))
            # Within recurrent layer, include dropout
        else:
            model.add(GRU(self.units,
                          input_shape=(X_train.shape[1], X_train.shape[2]),
                          dropout=self.dropout,
                          recurrent_dropout=self.dropout))
            # Within recurrent layer, include dropout
        if self.dropout != 0:
            # Dropout some units (recurrent layer output units)
            model.add(Dropout(self.dropout))

        # Add dense connections to output layer
        model.add(Dense(y_train.shape[1]))
        model.add(Activation('softplus'))

        # Fit model (and set fitting parameters)
        model.compile(loss='categorical_crossentropy', optimizer='rmsprop',
                      metrics=['accuracy'])  # Set loss function and optimizer
        if keras_v1:
            model.fit(X_train, y_train, nb_epoch=self.num_epochs,
                      verbose=self.verbose)  # Fit the model
        else:
            model.fit(X_train, y_train, epochs=self.num_epochs,
                      verbose=self.verbose)  # Fit the model
        self.model = model

    def predict(self, X_test):

        """Predict outcomes using trained LSTM Decoder

        Parameters
        ----------
        X_test: numpy 3d array of shape [n_samples,n_time_bins,n_neurons]
            This is the neural data being used to predict outputs.

        Returns
        -------
        y_test_predicted: numpy 2d array of shape [n_samples,n_outputs]
            The predicted outputs
        """

        y_test_predicted_raw = self.model.predict(X_test)  # Make predictions
        y_test_predicted = np.argmax(y_test_predicted_raw, axis=1)

        return y_test_predicted


# %% LONG SHORT TERM MEMORY (LSTM) DECODER

class LSTMClassification(object):
    """Class for the LSTM decoder

    Parameters
    ----------
    units: integer, optional, default 400
        Number of hidden units in each layer

    dropout: decimal, optional, default 0
        Proportion of units that get dropped out

    num_epochs: integer, optional, default 10
        Number of epochs used for training

    verbose: binary, optional, default=0
        Whether to show progress of the fit after each epoch
    """

    def __init__(self, units=400, dropout=0, num_epochs=10, verbose=0):
        self.units = units
        self.dropout = dropout
        self.num_epochs = num_epochs
        self.verbose = verbose

    def fit(self, X_train, y_train):

        """Train LSTM Decoder

        Parameters
        ----------
        X_train: numpy 3d array of shape [n_samples,n_time_bins,n_neurons]
            This is the neural data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 2d array of shape [n_samples, n_outputs]
            This is the outputs that are being predicted
        """

        # Use one-hot coding for y
        if y_train.ndim == 1:
            y_train = np_utils.to_categorical(y_train.astype(int))
        elif y_train.shape[1] == 1:
            y_train = np_utils.to_categorical(y_train.astype(int))

        model = Sequential()  # Declare model
        # Add recurrent layer
        if keras_v1:
            model.add(LSTM(self.units,
                           input_shape=(X_train.shape[1], X_train.shape[2]),
                           dropout_W=self.dropout,
                           dropout_U=self.dropout))
            # Within recurrent layer, include dropout
        else:
            model.add(LSTM(self.units,
                           input_shape=(X_train.shape[1], X_train.shape[2]),
                           dropout=self.dropout,
                           recurrent_dropout=self.dropout))
            # Within recurrent layer, include dropout
        if self.dropout != 0:
            # Dropout some units (recurrent layer output units)
            model.add(Dropout(self.dropout))

        # Add dense connections to output layer
        model.add(Dense(y_train.shape[1]))
        model.add(Activation('softplus'))

        # Fit model (and set fitting parameters)
        model.compile(loss='categorical_crossentropy', optimizer='rmsprop',
                      metrics=['accuracy'])  # Set loss function and optimizer
        if keras_v1:
            model.fit(X_train, y_train, nb_epoch=self.num_epochs,
                      verbose=self.verbose)  # Fit the model
        else:
            model.fit(X_train, y_train, epochs=self.num_epochs,
                      verbose=self.verbose)  # Fit the model
        self.model = model

    def predict(self, X_test):

        """Predict outcomes using trained LSTM Decoder

        Parameters
        ----------
        X_test: numpy 3d array of shape [n_samples,n_time_bins,n_neurons]
            This is the neural data being used to predict outputs.

        Returns
        -------
        y_test_predicted: numpy 2d array of shape [n_samples,n_outputs]
            The predicted outputs
        """

        y_test_predicted_raw = self.model.predict(X_test)  # Make predictions
        y_test_predicted = np.argmax(y_test_predicted_raw, axis=1)

        return y_test_predicted


# %% EXTREME GRADIENT BOOSTING (XGBOOST)

class XGBoostClassification(object):
    """Class for the XGBoost Decoder

    Parameters
    ----------
    max_depth: integer, optional, default=3
        the maximum depth of the trees

    num_round: integer, optional, default=300
        the number of trees that are fit

    eta: float, optional, default=0.3
        the learning rate

    gpu: integer, optional, default=-1
        if the gpu version of xgboost is installed, this can be used to select
        which gpu to use
        for negative values (default), the gpu is not used
    """

    def __init__(self, max_depth=3, num_round=300, eta=0.3, gpu=-1):
        self.max_depth = max_depth
        self.num_round = num_round
        self.eta = eta
        self.gpu = gpu

    def fit(self, X_flat_train, y_train):

        """Train XGBoost Decoder

        Parameters
        ----------
        X_flat_train: numpy 2d array of shape [n_samples,n_features]
            This is the neural data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 1d array of shape (n_samples), with integers
            representing classes or 2d array of shape [n_samples, n_outputs] in
             1-hot form. This is the outputs that are being predicted
        """

        # turn to categorial (not 1-hat)
        if (y_train.ndim == 2):
            if (y_train.shape[1] == 1):
                y_train = np.reshape(y_train, -1)
            else:
                y_train = np.argmax(y_train, axis=1, out=None)

        # Get number of classes
        n_classes = len(np.unique(y_train))

        # Set parameters for XGBoost
        param = {'objective': "multi:softmax",  # or softprob
                 'eval_metric': "mlogloss",  # loglikelihood loss
                 # 'eval_metric': "merror",
                 'max_depth': self.max_depth,
                 # this is the only parameter we have set, it's one of the way
                 # or regularizing
                 'eta': self.eta,
                 'num_class': n_classes,  # y_train.shape[1],
                 'seed': 2925,  # for reproducibility
                 'silent': 1}
        if self.gpu < 0:
            param['nthread'] = -1  # with -1 it will use all available threads
        else:
            param['gpu_id'] = self.gpu
            param['updater'] = 'grow_gpu'

        dtrain = xgb.DMatrix(X_flat_train,
                             label=y_train)  # Put in correct format for XGB
        bst = xgb.train(param, dtrain, self.num_round)  # Train model

        self.model = bst

    def predict(self, X_flat_test):

        """Predict outcomes using trained XGBoost Decoder

        Parameters
        ----------
        X_flat_test: numpy 2d array of shape [n_samples,n_features]
            This is the neural data being used to predict outputs.

        Returns
        -------
        y_test_predicted: numpy 1d array with integers as classes
            The predicted outputs
        """

        dtest = xgb.DMatrix(X_flat_test)  # Put in XGB format
        bst = self.model  # Get fit model
        y_test_predicted = bst.predict(dtest)  # Make prediction
        return y_test_predicted


# %% PRINCIPAL COMPONENT ANALYSIS - LINEAR DISCRIMINANT CLASSIFIER

class PcaLdaClassification(BaseEstimator):
    """Class for the PCA - LDA Classifier

    Parameters
    ----------
    explained variance: integer, optional, default=80
        the number of modes that explain the cumulative variance of the dataset

    da_type: string, optional, default=lda
        type of discriminant analysis; lda or qda

    """

    def __init__(self, explained_variance=0.8, da_type='lda', PCA_kwargs={},
                 DA_kwargs={}):
        # choose discriminant type
        if (da_type == 'lda'):
            # linear discriminant analysis
            da_model = da.LinearDiscriminantAnalysis(**DA_kwargs)
        else:
            # Quadratic discriminant analysis
            da_model = da.QuadraticDiscriminantAnalysis(**DA_kwargs)
        PCA_kwargs['n_components'] = explained_variance
        # Create a pipeline classifier
        self.model = Pipeline(steps=[
            ('pca', PCA(**PCA_kwargs)),
            ('discriminant', da_model)])

        # set default outputs to numpy
        self.model.set_output(transform="default")

    def fit(self, X_flat_train, y_train):

        """Train PCA - LDA classifier

        Parameters
        ----------
        X_flat_train: numpy 2d array of shape [n_samples,n_features]
            This is the neural data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 1d array of shape (n_samples), with integers
        representing classes
            This is the outputs that are being predicted
        """

        # Fit the model
        self.model.fit(X_flat_train, y_train)

    def predict(self, X_flat_test):

        """Predict outcomes using trained PCA LDA Decoder

        Parameters
        ----------
        X_flat_test: numpy 2d array of shape [n_samples,n_features]
            This is the neural data being used to predict outputs.

        Returns
        -------
        y_test_predicted: numpy 1d array with integers as classes
            The predicted outputs
        """

        pca_lda_fit = self.model  # Get fit model
        y_test_predicted = pca_lda_fit.predict(X_flat_test)  # Make prediction
        return y_test_predicted

    def get_scores(self, deep=True):
        """Get scores of pca and lda model

        Args:
            deep (bool, optional): Defaults to True.

        Returns:
            scores: dict
            Returns fitted scores of PCA and LDA
        """
        pca = self.model['pca']
        da = self.model['discriminant']
        scores = dict()
        scores['pca'] = pca.componens_
        scores['discriminant'] = da.coef_
        return scores

    def score(self, X, y, sample_weight=None):
        """Returns the mean accuracy on the given test data and labels.

        In multi-label classification, this is the subset accuracy
        which is a harsh metric since you require for each sample that
        each label set be correctly predicted.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples.

        y : array-like, shape = (n_samples) or (n_samples, n_outputs)
            True labels for X.

        sample_weight : array-like, shape = [n_samples], optional
            Sample weights.

        Returns
        -------
        score : float
            Mean accuracy of self.predict(X) wrt. y.

        """
        from sklearn.metrics import accuracy_score
        return accuracy_score(y, self.predict(X), sample_weight=sample_weight)

    # %% PRINCIPAL COMPONENT ANALYSIS Wrapper for classification function


class PcaEstimateDecoder(BaseEstimator):
    """Class for the PCA - SVM Classifier

    Parameters
    ----------
    explained_variance: float, optional, default=0.8
        the cumulative explained variance ratio required for PCA

    clf: object, optional, default=SVC()
        the classifier object to use for classification

    clf_params: dict, optional
        Additional parameters to be passed to the classifier (e.g.,
        {'param_name': value})

    """

    def __init__(self, explained_variance=0.8, clf=SVC(), clf_params=None):
        self.explained_variance = explained_variance
        self.clf = clf
        self.clf_params = clf_params
        self._initialize_model()

    def _initialize_model(self):
        self.pca = PCA(n_components=self.explained_variance)
        if self.clf_params is not None:

            self.model = Pipeline(steps=[('pca', self.pca), (
                'model', self.clf.set_params(**self.clf_params))])
        else:
            print('No initial parameters')
            self.model = Pipeline(
                steps=[('pca', self.pca), ('model', self.clf)])

    def fit(self, X_flat_train, y_train):
        """Train PCA - SVM classifier

        Parameters
        ----------
        X_flat_train: numpy 2d array of shape [n_samples, n_features]
            This is the neural data.
            See example file for an example of how to format the neural data
            correctly

        y_train: numpy 1d array of shape (n_samples), with integers
        representing classes
            This is the outputs that are being predicted
        """
        self._initialize_model()
        self.model.fit(X_flat_train, y_train)

    def predict(self, X_flat_test):
        """Predict outcomes using trained PCA - SVM Decoder

        Parameters
        ----------
        X_flat_test: numpy 2d array of shape [n_samples, n_features]
            This is the neural data being used to predict outputs.

        Returns
        -------
        y_test_predicted: numpy 1d array with integers as classes
            The predicted outputs
        """
        return self.model.predict(X_flat_test)

    def score(self, X, y, sample_weight=None):
        """Returns the mean accuracy on the given test data and labels.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples.

        y : array-like, shape = (n_samples,) or (n_samples, n_outputs)
            True labels for X.

        sample_weight : array-like, shape = [n_samples], optional
            Sample weights.

        Returns
        -------
        score : float
            Mean accuracy of self.predict(X) wrt. y.

        """
        return accuracy_score(y, self.predict(X), sample_weight=sample_weight)



from mne.utils import logger


def _qt_backend():
    """Set the backend to Qt5Agg"""
    import matplotlib
    try:
        matplotlib.use('Qt5Agg')
    except ImportError:
        logger.warn("Qt5Agg backend not available, using default backend")


_qt_backend()



import numpy as np
from ieeg.calc._fast.ufuncs import mean_diff as _md
from ieeg.calc._fast.mixup import mixupnd as cmixup, normnd as cnorm
from ieeg.calc._fast.permgt import permgtnd as permgt
from ieeg.calc._fast.concat import nan_concatinate

__all__ = ["mean_diff", "mixup", "permgt", "norm", "concatenate_arrays"]


def concatenate_arrays(arrays: tuple[np.ndarray, ...], axis: int = 0
                       ) -> np.ndarray:
    """Concatenate arrays along a specified axis, filling in empty arrays with
    nan values.

    Parameters
    ----------
    arrays
        A list of arrays to concatenate
    axis
        The axis along which to concatenate the arrays

    Returns
    -------
    result
        The concatenated arrays

    Examples
    --------
    >>> concatenate_arrays((np.array([1, 2, 3]), np.array([4, 5])), axis=None)
    array([[ 1.,  2.,  3.],
           [ 4.,  5., nan]])
    >>> concatenate_arrays((np.array([1, 2, 3]), np.array([4, 5])), axis=0)
    array([1., 2., 3., 4., 5.])
    >>> arr1 = np.arange(6, dtype=float).reshape(1, 2, 3)
    >>> arr2 = np.arange(24, dtype=float).reshape(2, 3, 4)
    >>> concatenate_arrays((arr1[0], arr2[0]), axis=0)
    array([[ 0.,  1.,  2., nan],
           [ 3.,  4.,  5., nan],
           [ 0.,  1.,  2.,  3.],
           [ 4.,  5.,  6.,  7.],
           [ 8.,  9., 10., 11.]])
    >>> concatenate_arrays((arr2[0], arr1[0]), axis=1)
    array([[ 0.,  1.,  2.,  3.,  0.,  1.,  2.],
           [ 4.,  5.,  6.,  7.,  3.,  4.,  5.],
           [ 8.,  9., 10., 11., nan, nan, nan]])
    >>> arr = concatenate_arrays((arr1[0], arr2[0]), axis=None)
    >>> arr
    array([[[ 0.,  1.,  2., nan],
            [ 3.,  4.,  5., nan],
            [nan, nan, nan, nan]],
    <BLANKLINE>
           [[ 0.,  1.,  2.,  3.],
            [ 4.,  5.,  6.,  7.],
            [ 8.,  9., 10., 11.]]])
    """

    if axis is None:
        axis = 0
        arrays = [np.expand_dims(ar, axis) for ar in arrays]

    arrays = [ar.astype(float) for ar in arrays if ar.size > 0]

    while axis < 0:
        axis += max(a.ndim for a in arrays)

    return nan_concatinate(arrays, axis)


def mixup(arr: np.ndarray, obs_axis: int, alpha: float = 1.,
          seed: int = None) -> None:
    """Oversample by mixing two random non-NaN observations

    Parameters
    ----------
    arr : array
        The data to oversample.
    obs_axis : int
        The axis along which to apply func.
    alpha : float
        The alpha parameter for the beta distribution. If alpha is 0, then
        the distribution is uniform. If alpha is 1, then the distribution is
        symmetric. If alpha is greater than 1, then the distribution is
        skewed towards the first observation. If alpha is less than 1, then
        the distribution is skewed towards the second observation.

    Examples
    --------
    >>> arr = np.array([[1, 2], [4, 5], [7, 8],
    ... [float("nan"), float("nan")]])
    >>> mixup(arr, 0, seed=42)
    >>> arr # doctest: +NORMALIZE_WHITESPACE +SKIP
    array([[1.        , 2.        ],
           [4.        , 5.        ],
           [7.        , 8.        ],
           [5.24946679, 6.24946679]])
    >>> arr2 = np.arange(24, dtype=float).reshape(2, 3, 4)
    >>> arr2[0, 2, :] = [float("nan")] * 4
    >>> mixup(arr2, 1, seed=42)
    >>> arr2 # doctest: +NORMALIZE_WHITESPACE +SKIP
    array([[[ 0.        ,  1.        ,  2.        ,  3.        ],
            [ 4.        ,  5.        ,  6.        ,  7.        ],
            [ 2.33404428,  3.33404428,  4.33404428,  5.33404428]],
    <BLANKLINE>
           [[12.        , 13.        , 14.        , 15.        ],
            [16.        , 17.        , 18.        , 19.        ],
            [20.        , 21.        , 22.        , 23.        ]]])
    >>> arr3 = np.arange(24, dtype=float).reshape(3, 2, 4)
    >>> arr3[0, :, :] = float("nan")
    >>> mixup(arr3, 0, seed=42)
    >>> arr3 # doctest: +NORMALIZE_WHITESPACE +SKIP
    array([[[12.66808855, 13.66808855, 14.66808855, 15.66808855],
            [17.31717879, 18.31717879, 19.31717879, 20.31717879]],
    <BLANKLINE>
           [[ 8.        ,  9.        , 10.        , 11.        ],
            [12.        , 13.        , 14.        , 15.        ]],
    <BLANKLINE>
           [[16.        , 17.        , 18.        , 19.        ],
            [20.        , 21.        , 22.        , 23.        ]]])
    """
    if seed is None:
        seed = np.random.randint(0, 2**16 - 1)

    cmixup(arr, obs_axis, alpha, seed)


def norm(arr: np.ndarray, obs_axis: int = -1) -> None:
    """Oversample by obtaining the distribution and randomly selecting

    Parameters
    ----------
    arr : array
        The data to oversample.
    obs_axis : int
        The axis along which to apply func.

    Examples
    --------
    >>> np.random.seed(0)
    >>> arr = np.array([1, 2, 4, 5, 7, 8,
    ... float("nan"), float("nan")])
    >>> norm(arr)
    >>> arr
    array([1.        , 2.        , 4.        , 5.        , 7.        ,
           8.        , 8.91013086, 5.50039302])
    """
    cnorm(arr, obs_axis)


def mean_diff(group1: np.ndarray, group2: np.ndarray,
              axis: int = -1) -> np.ndarray | float:
    """Calculate the mean difference between two groups.

    This function is the default statistic function for time_perm_cluster. It
    calculates the mean difference between two groups along the specified axis.

    Parameters
    ----------
    group1 : array, shape (..., time)
        The first group of observations.
    group2 : array, shape (..., time)
        The second group of observations.
    axis : int or tuple of ints, optional
        The axis or axes along which to compute the mean difference. If None,
        compute the mean difference over all axes.

    Returns
    -------
    avg1 - avg2 : array or float
        The mean difference between the two groups.

    Examples
    --------
    >>> import numpy as np
    >>> group1 = np.array([[1, 1, 1, 1, 1], [0, 60, 0, 10, 0]], order='F').T
    >>> group2 = np.array([[1, 1, 1, 1, 1], [0, 0, 0, 0, 0]], order='F').T
    >>> mean_diff(group1, group2, axis=0)
    array([ 0., 14.])
    >>> mean_diff(group1, group2, axis=1)
    array([ 0., 30.,  0.,  5.,  0.])
    """
    in1 = np.moveaxis(group1, axis, -1)
    in2 = np.moveaxis(group2, axis, -1)

    return _md(in1, in2)



import functools
from collections.abc import Iterable

import mne
from ieeg.calc.fast import concatenate_arrays

import numpy as np
from numpy.matlib import repmat
from numpy.typing import ArrayLike
from numpy.core.numeric import normalize_axis_tuple

import ieeg


def iter_nest_dict(d: dict, _lvl: int = 0, _coords=()):
    """Iterate over a nested dictionary, yielding the key and value.

    Parameters
    ----------
    d : dict
        The dictionary to iterate over.

    Yields
    ------
    tuple
        The key and value of the dictionary.

    Examples
    --------
    >>> d = {'a': {'b': 1, 'c': 2}, 'd': {'e': 3, 'f': 4}}
    >>> for k, v in iter_nest_dict(d):
    ...     print(k, v)
    ('a', 'b') 1
    ('a', 'c') 2
    ('d', 'e') 3
    ('d', 'f') 4
    """
    for k, v in d.items():
        if isinstance(v, dict):
            yield from iter_nest_dict(v, _lvl + 1, _coords + (k,))
        elif isinstance(v, np.ndarray):
            yield from iter_nest_dict({i: val for i, val in enumerate(v)
                                       }, _lvl + 1, _coords + (k,))
        else:
            yield _coords + (k,), v


class LabeledArray(np.ndarray):
    """ A numpy array with labeled dimensions, acting like a dictionary.

    A numpy array with labeled dimensions. This class is useful for storing
    data that is not easily represented in a tabular format. It acts as a
    nested dictionary but its values map to elements of a stored numpy array.

    Parameters
    ----------
    input_array : array_like
        The array to store in the LabeledArray.
    labels : tuple[tuple[str, ...], ...], optional
        The labels for each dimension of the array, by default ().
    delimiter : str, optional
        The delimiter to use when combining labels, by default '-'
    **kwargs
        Additional arguments to pass to np.asarray.

    Attributes
    ----------
    labels : tuple[tuple[str, ...], ...]
        The labels for each dimension of the array.
    array : np.ndarray
        The array stored in the LabeledArray.

    Examples
    --------
    >>> import numpy as np
    >>> from ieeg.calc.mat import LabeledArray
    >>> arr = np.ones((2, 3, 4), dtype=int)
    >>> labels = (('a', 'b'), ('c', 'd', 'e'), ('f', 'g', 'h', 'i'))
    >>> la = LabeledArray(arr, labels)
    >>> la
    array([[[1, 1, 1, 1],
            [1, 1, 1, 1],
            [1, 1, 1, 1]],
    <BLANKLINE>
           [[1, 1, 1, 1],
            [1, 1, 1, 1],
            [1, 1, 1, 1]]])
    labels(['a', 'b']
           ['c', 'd', 'e']
           ['f', 'g', 'h', 'i'])
    >>> la.to_dict() # doctest: +ELLIPSIS
    {'a': {'c': {'f': 1, 'g': 1, 'h': 1, 'i': 1}, 'd': {'f': 1, 'g': 1,...
    >>> la['a', 'c', 'f'] = 2
    >>> la['a', 'c', 'f']
    2
    >>> la['a', 'c']
    array([2, 1, 1, 1])
    labels(['f', 'g', 'h', 'i'])
    >>> la['a'].labels
    [['c', 'd', 'e'], ['f', 'g', 'h', 'i']]
    >>> la['a','d'] = np.array([3,3,3,3])
    >>> la[('a','b'), :]
    array([[[2, 1, 1, 1],
            [3, 3, 3, 3],
            [1, 1, 1, 1]],
    <BLANKLINE>
           [[1, 1, 1, 1],
            [1, 1, 1, 1],
            [1, 1, 1, 1]]])
    labels(['a', 'b']
           ['c', 'd', 'e']
           ['f', 'g', 'h', 'i'])
    >>> la[np.array([False, True]),]
    array([[[1, 1, 1, 1],
            [1, 1, 1, 1],
            [1, 1, 1, 1]]])
    labels(['b']
           ['c', 'd', 'e']
           ['f', 'g', 'h', 'i'])
    >>> la[(0, 1)]
    array([3, 3, 3, 3])
    labels(['f', 'g', 'h', 'i'])
    >>> la[0, 1]
    array([3, 3, 3, 3])
    labels(['f', 'g', 'h', 'i'])
    >>> la[(0, 1),].labels
    [['a', 'b'], ['c', 'd', 'e'], ['f', 'g', 'h', 'i']]
    >>> np.nanmean(la, axis=(-2, -1))
    array([1.75, 1.  ])
    labels(['a', 'b'])
    >>> arr = np.arange(24).reshape((2, 3, 4))
    >>> labels = (('a', 'b'), ('c', 'd', 'e'), ('f', 'g', 'h', 'i'))
    >>> ad = LabeledArray(arr, labels)
    >>> ad[None, 'a'].labels
    [['1'], ['c', 'd', 'e'], ['f', 'g', 'h', 'i']]
    >>> ad['b', 0, np.array([[1,2], [0,3]])]
    array([[13, 14],
           [12, 15]])
    labels(['g-h', 'f-i']
           ['f-g', 'h-i'])
    >>> ad[:, ('d','e'),][..., ('g', 'h'),].labels
    [['a', 'b'], ['d', 'e'], ['g', 'h']]
    >>> ad['a', 'd', ('g', 'i', 'f'),]
    array([5, 7, 4])
    labels(['g', 'i', 'f'])

    Notes
    -----
    Multiple sequence advanced indices objects are not supported. If you want
     to use multiple sequence indices, you should use them one at a time.

    References
    ----------
    [1] https://numpy.org/doc/stable/user/basics.subclassing.html
    [2] https://numpy.org/doc/stable/user/basics.indexing.html
    """

    labels: list = []

    def __new__(cls, input_array, labels: list[tuple[str, ...], ...] = (),
                delimiter: str = '-', **kwargs):
        obj = np.asarray(input_array, **kwargs).view(cls)
        labels = list(labels)
        for i in range(obj.ndim):
            if len(labels) < i + 1:
                labels.append(tuple(range(obj.shape[i])))
        obj.labels = list(map(lambda lab: Labels(lab, delimiter), labels))
        assert tuple(map(len, obj.labels)) == obj.shape, \
            f"labels must have the same length as the shape of the array, " \
            f"instead got {tuple(map(len, obj.labels))} and {obj.shape}"
        return obj

    def __array_finalize__(self, obj, *args, **kwargs):
        if obj is None:
            return
        self.labels = getattr(obj, 'labels', kwargs.pop('labels', ()))
        super(LabeledArray, self).__array_finalize__(obj, *args, **kwargs)

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(LabeledArray, self).__reduce__()
        # Create our own tuple to pass to __setstate__
        new_state = pickled_state[2] + (self.labels,)
        # Return a tuple that replaces the parent's __setstate__
        # tuple with our own
        return (pickled_state[0], pickled_state[1], new_state)

    def __setstate__(self, state):
        self.labels = state[-1]  # Set the info attribute
        # Call the parent's __setstate__ with the other tuple elements.
        super(LabeledArray, self).__setstate__(state[0:-1])

    def __array_ufunc__(self, ufunc, method, *inputs, out=None, **kwargs):
        la_inputs = (i for i in inputs if isinstance(i, LabeledArray))
        labels = next(la_inputs).labels.copy()
        inputs = tuple(i.view(np.ndarray) if isinstance(i, LabeledArray)
                       else i for i in inputs)
        if out is not None:
            kwargs['out'] = tuple(o.view(np.ndarray) if
                                  isinstance(o, LabeledArray)
                                  else o for o in out)
        if method == 'reduce':
            axis = kwargs.get('axis', None)
            if axis is None:
                axis = range(inputs[0].ndim)
            elif np.isscalar(axis):
                axis = (axis,)
            else:
                axis = tuple(axis)
            i = 0
            for ax in axis:
                if ax > 0:
                    ax -= i
                labels = list(labels)
                if kwargs.get('keepdims', False):
                    labels[ax] = ("-".join(labels[ax]),)
                else:
                    labels.pop(ax)
                    i += 1
                labels = tuple(labels)

        outputs = super(LabeledArray, self).__array_ufunc__(
            ufunc, method, *inputs, **kwargs)
        if isinstance(outputs, tuple):
            outputs = tuple(LabeledArray(o, labels)
                            if isinstance(o, np.ndarray)
                            else o for o in outputs)
        elif isinstance(outputs, np.ndarray):
            outputs = LabeledArray(outputs, labels)
        return outputs

    @property
    def T(self):
        return LabeledArray(self.__array__().T, self.labels[::-1])

    def swapaxes(self, axis1, axis2):
        new = list(self.labels)
        new[axis1], new[axis2] = new[axis2], new[axis1]
        arr = super(LabeledArray, self).swapaxes(axis1, axis2)
        return LabeledArray(arr, new)

    def transpose(self, axes):
        axes = normalize_axis_tuple(axes, self.ndim)
        new_labels = [self.labels[i] for i in axes]
        arr_t = super(LabeledArray, self).transpose(axes)
        return LabeledArray(arr_t, new_labels)

    @classmethod
    def from_dict(cls, data: dict, **kwargs) -> 'LabeledArray':
        """Create a LabeledArray from a dictionary.

        Parameters
        ----------
        data : dict
            The dictionary to convert to a LabeledArray.

        Returns
        -------
        LabeledArray
            The LabeledArray created from the dictionary.

        Examples
        --------
        >>> data = {'a': {'b': {'c': 1}}}
        >>> LabeledArray.from_dict(data, dtype=int) # doctest: +ELLIPSIS
        array([[[1]]])
        labels(['a']
               ['b']
               ['c'])
        >>> data = {'a': {'b': {'c': 1}}, 'd': {'b': {'c': 2, 'e': 3}}}
        >>> LabeledArray.from_dict(data) # doctest: +ELLIPSIS
        array([[[ 1., nan]],
        <BLANKLINE>
               [[ 2.,  3.]]])
        labels(['a', 'd']
               ['b']
               ['c', 'e'])
        """

        arr = inner_array(data)
        keys = inner_all_keys(data)
        return cls(arr, keys, **kwargs)

    @classmethod
    def from_signal(cls, sig: ieeg.Signal, **kwargs) -> 'LabeledArray':
        """Create a LabeledArray from a Signal.

        Parameters
        ----------
        sig : Signal
            The Signal to convert to a LabeledArray.

        Returns
        -------
        LabeledArray
            The LabeledArray created from the Signal.

        Examples
        --------
        >>> from bids import BIDSLayout
        >>> from ieeg.io import raw_from_layout
        >>> from ieeg.navigate import trial_ieeg
        >>> import sys
        >>> bids_root = mne.datasets.epilepsy_ecog.data_path()
        >>> layout = BIDSLayout(bids_root)
        >>> with mne.use_log_level(0):
        ...     raw = raw_from_layout(layout, subject="pt1", preload=True,
        ...     extension=".vhdr", verbose=False)
        >>> LabeledArray.from_signal(raw, dtype=float) # doctest: +ELLIPSIS
        array([[-8.98329883e-06,  8.20419238e-06,  7.42294287e-06, ...,
                 1.07177293e-09,  1.07177293e-09,  1.07177293e-09],
               [ 2.99222000e-04,  3.03518844e-04,  2.96878250e-04, ...,
                 3.64667153e-09,  3.64667153e-09,  3.64667153e-09],
               [ 2.44140953e-04,  2.30078469e-04,  2.19140969e-04, ...,
                 3.85053724e-10,  3.85053724e-10,  3.85053724e-10],
               ...,
               [ 1.81263844e-04,  1.74232594e-04,  1.56263875e-04, ...,
                 1.41283798e-08,  1.41283798e-08,  1.41283798e-08],
               [ 2.25390219e-04,  2.16015219e-04,  1.91405859e-04, ...,
                -2.91418821e-10, -2.91418821e-10, -2.91418821e-10],
               [ 3.14092313e-04,  3.71123375e-04,  3.91826437e-04, ...,
                 3.07457047e-08,  3.07457047e-08,  3.07457047e-08]])
        labels(['G1', 'G2', 'G3', 'G4', 'G5', 'G6', 'G7', 'G8', 'G9', ...
        >>> epochs = trial_ieeg(raw, "AD1-4, ATT1,2", (-1, 2), preload=True,
        ... verbose=False)
        >>> LabeledArray.from_signal(epochs, dtype=float) # doctest: +ELLIPSIS
        array([[[ 0.00021563,  0.00021563,  0.00020703, ..., -0.00051211,
                 -0.00051445, -0.00050351],
                [-0.00030586, -0.00030625, -0.00031171, ..., -0.00016054,
                 -0.00015976, -0.00015664],
                [-0.00010781, -0.00010469, -0.00010859, ...,  0.00026719,
                  0.00027695,  0.00030156],
                ...,
                [-0.00021483, -0.00021131, -0.00023084, ..., -0.00034295,
                 -0.00032381, -0.00031444],
                [-0.00052188, -0.00052852, -0.00053125, ..., -0.00046211,
                 -0.00047148, -0.00047891],
                [-0.00033708, -0.00028005, -0.00020934, ..., -0.00040934,
                 -0.00042341, -0.00040973]]])
        ...

        """

        arr = sig.get_data()
        match sig:
            case mne.io.base.BaseRaw():
                labels = [sig.ch_names, sig.times]
            case mne.BaseEpochs():
                events = events_in_order(sig)
                labels = [events, sig.ch_names, sig.times]
            case mne.evoked.Evoked():
                labels = [sig.ch_names, sig.times]
            case mne.time_frequency.EpochsTFR():
                events = events_in_order(sig)
                labels = [events, sig.ch_names, sig.freqs, sig.times]
            case mne.time_frequency.AverageTFR():
                labels = [sig.ch_names, sig.freqs, sig.times]
            case _:
                raise TypeError(f"Unexpected data type: {type(sig)}")
        return cls(arr, labels, **kwargs)

    def _parse_index(self, keys: list) -> list:
        ndim = self.ndim
        new_keys = [range(self.shape[i]) for i in range(ndim)]
        dim = 0
        newaxis_count = 0
        for i, key in enumerate(keys):
            key_type = type(key)
            if np.issubdtype(key_type, str):
                key = self.labels[dim - newaxis_count].find(key)
                keys[i] = key  # set original keys as well
            elif key is Ellipsis:
                num_ellipsis_dims = ndim - len(keys) + 1
                while dim < num_ellipsis_dims:
                    dim += 1
                continue
            elif key_type is slice:
                key = new_keys[dim][key]
            elif key is np.newaxis or key is None:
                new_keys.insert(dim, None)
                newaxis_count += 1
                dim += 1
                continue
            elif (key_type in (list, tuple) or
                  np.issubdtype(key_type, np.ndarray)):
                key = list(key)
                for j, k in enumerate(key):
                    if np.issubdtype(type(k), str):
                        key[j] = self.labels[dim - newaxis_count].find(k)
                if np.issubdtype(key_type, np.ndarray):
                    keys[i] = np.array(key)
                else:
                    keys[i] = key_type(key)
            elif np.isscalar(key):  # key should be an int
                while key < 0:
                    key += self.shape[dim - newaxis_count]
            else:
                raise TypeError(f"Unexpected key type: {key_type}")

            new_keys[dim] = key
            dim += 1
        return new_keys

    def _to_coords(self, orig_keys):

        if np.isscalar(orig_keys) or np.issubdtype(
                (dtype := getattr(orig_keys, 'dtype', None)), np.integer):
            keys = [orig_keys]
            l_keys = self._parse_index(keys)
            return keys[0], tuple(l_keys)
        elif dtype == np.bool_ and is_broadcastable(
                getattr(orig_keys, 'shape', ()), self.shape):
            l_keys = np.where(np.reshape(orig_keys, self.shape))
            return orig_keys, l_keys
        else:
            if isinstance(orig_keys, slice):
                keys = [orig_keys]
            else:
                keys = list(orig_keys)
            l_keys = self._parse_index(keys)
            return tuple(keys), tuple(l_keys)

    def __getitem__(self, orig_keys):
        keys, label_keys = self._to_coords(orig_keys)
        out = super(LabeledArray, self).__getitem__(keys)
        if out.ndim == 0:
            return out[()]

        # determine the new labels
        new_labels = [None] * out.ndim
        j = 0
        k = 0
        for i, label_key in enumerate(label_keys):

            if label_key is None:
                new_labels[i - k] = Labels(['1'])
                j += 1
            elif np.isscalar(label_key):  # basic indexing triggered
                k += 1
            elif i - k >= out.ndim:
                raise IndexError(f"Too many indices for array: "
                                 f"array is {out.ndim}-dimensional, "
                                 f"but {i + 1} were indexed")
            else:
                if isinstance(label_key, tuple):
                    label_key = np.asarray(label_key)
                labels = np.atleast_1d(np.squeeze(self.labels[i - j][label_key]
                                                  ))
                if labels.ndim > 1:
                    lab_list = labels.decompose()
                    new_labels[i - k:i - k + len(labels)] = lab_list
                    k += len(lab_list) - 1
                else:
                    new_labels[i - k] = labels

        if any(l_none := lab is None for lab in new_labels):
            raise IndexError(f"Too few indices for array: array is {out.ndim}"
                             f"-dimensional, but {sum(~l_none)} were indexed")

        setattr(out, 'labels', new_labels)
        return out

    def __setitem__(self, keys, value):
        keys, _ = self._to_coords(keys)
        super(LabeledArray, self).__setitem__(keys, value)

    def __repr__(self):
        return repr(self.__array__()) + f"\nlabels({self._label_formatter()})"

    def __str__(self):
        return str(self.__array__()) + f"\nlabels({self._label_formatter()})"

    def _label_formatter(self):
        def _liststr(x):
            return f"\n       ".join(x)

        return _liststr([str(lab) for lab in self.labels])

    def memory(self):
        size = self.nbytes
        for unit in ['B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB']:
            if size < 1024.0 or unit == 'PiB':
                break
            size /= 1024.0
        return size, unit

    def __eq__(self, other):
        if isinstance(other, LabeledArray):
            return np.array_equal(self, other, True) and \
                all(np.array_equal(l1, l2) for l1, l2 in zip(self.labels,
                                                             other.labels))
        else:
            return self.__array__().__eq__(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_dict(self) -> dict:
        """Convert to a dictionary."""
        out = {}
        for k, v in self.items():
            if len(self.labels) > 1:
                out[k] = v.to_dict()
            elif np.isnan(v).all():
                continue
            else:
                out[k] = v
        return out

    def items(self):
        return zip(self.keys(), self.values())

    def keys(self):
        return (lab for lab in self.labels[0])

    def values(self):
        return (a for a in self)

    def _reshape(self, shape, order='C') -> 'LabeledArray':
        """Reshape the array.

        Parameters
        ----------
        shape : tuple[int, ...]
            The new shape of the array.
        order : str, optional
            The order to reshape the array in, by default 'C'

        Returns
        -------
        LabeledArray
            The reshaped LabeledArray.

        Examples
        --------
        >>> data = {'a': {'b': {'c': 1}}}
        >>> ad = LabeledArray.from_dict(data, dtype=int)
        >>> ad.labels
        [['a'], ['b'], ['c']]
        >>> ad._reshape((1, 1, 1)) # doctest: +SKIP
        array([[[1]]])
        labels(['a']
               ['b']
               ['c'])
        >>> arr = np.arange(24).reshape((2, 3, 4))
        >>> labels = [('a', 'b'), ('c', 'd', 'e'), ('f', 'g', 'h', 'i')]
        >>> ad = LabeledArray(arr, labels)
        >>> ad._reshape((6, 4))
        array([[ 0,  1,  2,  3],
               [ 4,  5,  6,  7],
               [ 8,  9, 10, 11],
               [12, 13, 14, 15],
               [16, 17, 18, 19],
               [20, 21, 22, 23]])
        labels(['a-c', 'a-d', 'a-e', 'b-c', 'b-d', 'b-e']
               ['f', 'g', 'h', 'i'])
        >>> ad._reshape((6, 4), 'F').labels
        [['a-c', 'b-c', 'a-d', 'b-d', 'a-e', 'b-e'], ['f', 'g', 'h', 'i']]
        >>> ad._reshape((2, 12)).labels # doctest: +ELLIPSIS
        [['a', 'b'], ['c-f', 'c-g', 'c-h', 'c-i', 'd-f', 'd-g', 'd-h', 'd-i'...
        >>> arr = np.arange(10)
        >>> labels = [list(map(str, arr))]
        >>> ad = LabeledArray(arr, labels)
        >>> ad._reshape((2, 5)).labels
        [['0-1-2-3-4', '5-6-7-8-9'], ['0-5', '1-6', '2-7', '3-8', '4-9']]
        >>> ad._reshape((1, 2, 5)).labels # doctest: +ELLIPSIS
        [['0-1-2-3-4-5-6-7-8-9'], ['0-1-2-3-4', '5-6-7-8-9'], ['0-5', '1-6',...
        """
        new_array = super(LabeledArray, self).reshape(*shape, order=order)
        lab_mat = functools.reduce(lambda x, y: x @ y, self.labels)
        new_labels = lab_mat.reshape(*shape, order=order).decompose()
        return LabeledArray(new_array, new_labels)

    def combine(self, levels: tuple[int, int]) -> 'LabeledArray':
        """Combine any levels of a LabeledArray into the lower level

        Takes the input LabeledArray and rearranges its dimensions.

        Parameters
        ----------
        levels : tuple[int, int]
            The levels to combine, e.g. (0, 1) will combine the 1st and 2nd
            level of the array labels into one level at the 2nd level.
        delim : str, optional
            The delimiter to use when combining labels, by default '-'

        Returns
        -------
        LabeledArray
            The combined LabeledArray

        Examples
        --------
        >>> data = {'a': {'b': {'c': 1}}}
        >>> ad = LabeledArray.from_dict(data, dtype=int)
        >>> ad.combine((0, 2))
        array([[1]])
        labels(['b']
               ['a-c'])
        >>> ad2 = LabeledArray([[[1,2],[3,4]],[[5,6],[7,8]]],
        ... labels=[('a', 'b'), ('c', 'd'), ('e', 'f')])
        >>> ad2['a', : , 'e']
        array([1, 3])
        labels(['c', 'd'])
        >>> ad2.combine((0, 2))
        array([[1, 2, 5, 6],
               [3, 4, 7, 8]])
        labels(['c', 'd']
               ['a-e', 'a-f', 'b-e', 'b-f'])
        >>> np.mean(ad2.combine((0, 2)), axis=1)
        array([3.5, 5.5])
        labels(['c', 'd'])
        >>> np.mean(ad2, axis=(0, 2))
        array([3.5, 5.5])
        labels(['c', 'd'])
        """

        assert levels[0] >= 0, "first level must be >= 0"
        assert levels[1] > levels[0], "second level must be > first level"

        new_labels = list(self.labels).copy()
        new_labels.pop(levels[0])

        new_labels[levels[1] - 1] = (
                self.labels[levels[0]] @ self.labels[levels[1]]).flatten()

        all_idx = ([slice(None) if i != levels[0] else sl for i in
                    range(self.ndim)] for sl in range(self.shape[levels[0]]))

        arrs = [self.__array__()[*idx] for idx in all_idx]
        new_array = concatenate_arrays(arrs, axis=levels[1] - 1)

        return LabeledArray(new_array, new_labels, dtype=self.dtype)

    def take(self, indices, axis=None, **kwargs):
        """Take elements from an array along an axis.

        This function does not support the out argument.

        Parameters
        ----------
        indices : array_like
            The indices of the values to extract.
        axis : int, optional
            The axis over which to select values, by default None.
        kwargs : dict
            Additional keyword arguments to pass to np.take.

        Examples
        --------
        >>> arr = np.arange(24).reshape((2, 3, 4))
        >>> labels = [('a', 'b'), ('c', 'd', 'e'), ('f', 'g', 'h', 'i')]
        >>> ad = LabeledArray(arr, labels)
        >>> ad.take([0, 2], axis=1).labels
        [['a', 'b'], ['c', 'e'], ['f', 'g', 'h', 'i']]
        >>> np.take_along_axis(ad, np.array([[[0, 1]]]), axis=2).labels
        [['a', 'b'], ['c', 'd', 'e'], ['f', 'g']]
        >>> np.take(ad, np.array([[0,2], [1,3]]), axis=2).labels
        [['a', 'b'], ['c', 'd', 'e'], ['f-h', 'g-i'], ['f-g', 'h-i']]
        """

        idx = [slice(None)] * self.ndim

        if axis is None:
            return self.flat[indices]
        elif isinstance(axis, int):
            idx[axis] = indices
        elif len(indices) == len(axis):
            for i, ax in enumerate(axis):
                idx[ax] = indices[i]
        else:
            raise ValueError("indices and axis must have the same length")

        out = super(LabeledArray, self).take(indices, axis, **kwargs)
        labels = [l[i] for i, l in zip(idx, self.labels)
                  if not np.isscalar(l[i])]
        for i, l in enumerate(labels):
            if l.ndim > 1:
                labels = labels[:i] + l.decompose()
        return LabeledArray(out, labels, dtype=self.dtype)

    def dropna(self) -> 'LabeledArray':
        """Remove all nan values from the array.

        Scans each column along any axis and removes all rows that contain
        only nan values.

        Returns
        -------
        LabeledArray
            The array with all nan values removed.

        Examples
        --------
        >>> data = {'a': {'b': {'c': 1., 'd': np.nan}}}
        >>> ad = LabeledArray.from_dict(data)
        >>> ad.dropna()
        array([[[1.]]])
        labels(['a']
               ['b']
               ['c'])
        >>> ad2 = LabeledArray([[[1,2],[3,4]],[[4,5],[6,7]],
        ... [[np.nan, np.nan], [np.nan, np.nan]]])
        >>> ad2.dropna()
        array([[[1., 2.],
                [3., 4.]],
        <BLANKLINE>
               [[4., 5.],
                [6., 7.]]])
        labels(['0', '1']
               ['0', '1']
               ['0', '1'])
        """
        new_labels = list(self.labels)
        idx = []
        for i in range(self.ndim):
            axes = tuple(j for j in range(self.ndim) if j != i)
            mask = np.all(np.isnan(np.array(self)), axis=axes)
            if np.any(mask):
                new_labels[i] = tuple(np.array(new_labels[i])[~mask])
            idx.append(~mask)
        index = np.ix_(*idx)
        return self[index]

    def concatenate(self, other: 'LabeledArray', axis: int = 0,
                    mismatch: str = 'raise', **kwargs) -> 'LabeledArray':
        """Concatenate two LabeledArrays along an axis.

        Parameters
        ----------
        other : LabeledArray
            The LabeledArray to concatenate with.
        axis : int, optional
            The axis to concatenate along, by default 0.
        mismatch : str, optional
            What to do if the number of labels are not the same, 'raise'
            (default) will raise a ValueError, 'shrink' will shrink the labels
            to the smallest size, and 'expand' (not implemented) will expand
            the labels to the largest size, filling in with NaNs.
        kwargs : dict
            Additional keyword arguments to pass to np.concatenate.

        Returns
        -------
        LabeledArray
            The concatenated LabeledArray.

        Examples
        --------
        >>> arr1 = LabeledArray([[1, 2],[3, 4]],
        ... labels=[('a', 'b'), ('c', 'd')])
        >>> arr2 = LabeledArray([[5, 6],[7, 8]],
        ... labels=[('a', 'b'), ('c', 'd')])
        >>> arr1.concatenate(arr2, axis=0)
        array([[1, 2],
               [3, 4],
               [5, 6],
               [7, 8]])
        labels(['a-0', 'b-0', 'a-1', 'b-1']
               ['c', 'd'])
        >>> arr3 = LabeledArray([[5, 6, 9],[7, 8, 10]],
        ... labels=[('a', 'b'), ('c', 'd', 'e')])
        >>> arr4 = LabeledArray([[1, 2, 3],[3, 4, 5]],
        ... labels=[('a', 'b'), ('c', 'e', 'd')])
        >>> arr3.concatenate(arr4, axis=0)
        array([[ 5,  6,  9],
               [ 7,  8, 10],
               [ 1,  3,  2],
               [ 3,  5,  4]])
        labels(['a-0', 'b-0', 'a-1', 'b-1']
               ['c', 'd', 'e'])
        >>> arr2.concatenate(arr4, axis=0) # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        ValueError: When mismatch is 'raise', the base array must the same s...
        >>> arr2.concatenate(arr4, 0, mismatch='shrink')
        array([[5, 6],
               [7, 8],
               [1, 3],
               [3, 5]])
        labels(['a-0', 'b-0', 'a-1', 'b-1']
               ['c', 'd'])
        >>> arr3.concatenate(arr1, 0, mismatch='shrink') # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        NotImplementedError: Base array must the same size or smaller than i...
        Base size:(2, 3), Input size: (2, 2)
        """

        while axis < 0:
            axis += self.ndim

        new_labels = list(self.labels)
        idx = [slice(None)] * self.ndim
        new = np.hstack((self.labels[axis], other.labels[axis]))
        for i in range(self.ndim):
            if i == axis:
                if not is_unique(new):
                    new_labels[i] = _make_array_unique(
                        new.astype(str), self.labels[i].delimiter)
                else:
                    new_labels[i] = new
            elif not (is_unique(new_labels[i]) and is_unique(other.labels[i])):
                raise NotImplementedError(
                    "Cannot concatenate arrays with non-unique labels "
                    f"{new_labels[i]}, {other.labels[i]}")
            elif self.shape[i] == other.shape[i]:
                if np.any(self.labels[i] != other.labels[i]):
                    idx[i] = get_subset_reorder_indices(
                        other.labels[i], self.labels[i])
            elif mismatch == 'raise':
                raise ValueError(
                    "When mismatch is 'raise', the base array must the same "
                    "size as the input array in all but the concatination "
                    f"axis, but along dimension {i} the base array has size "
                    f"{self.shape[i]} and the input array has size "
                    f"{other.shape[i]}")
            elif self.labels[i].shape[0] < other.labels[i].shape[0]:
                if mismatch == 'shrink':
                    idx[i] = get_subset_reorder_indices(
                        other.labels[i], self.labels[i])
                else:
                    raise NotImplementedError(
                        f"No method associated with mismatch = '{mismatch}',"
                        " try setting mismatch to 'shrink' or 'raise'")
            elif self.labels[i].shape[0] > other.labels[i].shape[0]:
                raise NotImplementedError(
                    "Base array must the same size or smaller than input "
                    "array in all but the concatination axes. \nBase size:"
                    f"{self.shape}, Input size: {other.shape}")
            else:
                raise ValueError("Unexpected error")

        reordered = other.__array__()[tuple(idx)]
        out = np.concatenate((self.__array__(), reordered), axis, **kwargs)
        return LabeledArray(out, new_labels, dtype=self.dtype)


def is_unique(arr: np.ndarray) -> bool:
    """Check if an array is unique.

    Parameters
    ----------
    arr : np.ndarray
        The array to check.

    Returns
    -------
    bool
        Whether the array is unique.

    Examples
    --------
    >>> is_unique(np.array([1, 2, 3]))
    True
    >>> is_unique(np.array([1, 2, 2]))
    False
    """
    return np.unique(arr).shape[0] == np.prod(arr.shape)


class Labels(np.char.chararray):
    """A class for storing labels for a LabeledArray."""
    delimiter: str

    # __slots__ = ['delimiter', '__dict__']

    def __new__(cls, input_array: ArrayLike, delim: str = '-'):
        obj = np.asarray(input_array, dtype=str).view(cls)
        setattr(obj, 'delimiter', delim)
        return obj

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(Labels, self).__reduce__()
        # Create our own tuple to pass to __setstate__
        new_state = pickled_state[2] + (self.delimiter,)
        # Return a tuple that replaces the parent's __setstate__ tuple with our
        # own
        return (pickled_state[0], pickled_state[1], new_state)

    def __setstate__(self, state):
        self.delimiter = state[-1]  # Set the info attribute
        # Call the parent's __setstate__ with the other tuple elements.
        super(Labels, self).__setstate__(state[0:-1])

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.delimiter = getattr(obj, 'delimiter', '-')

    def __str__(self):
        return self.tolist().__str__()

    def __repr__(self):
        return self.tolist().__repr__()

    def __matmul__(self, other):
        if not isinstance(other, Labels):
            raise NotImplementedError("Only Labels @ Labels is supported")
        s_str, o_str = self.astype(str), other.astype(str)

        # Convert the arrays to 2D
        s_str_2d = s_str[..., None]
        o_str_2d = o_str[None]

        # Use broadcasting to create a result array with combined strings
        result = s_str_2d + o_str_2d
        return result

    def __add__(self, other):
        result = super().__add__(self.delimiter).__add__(other)
        return Labels(result)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if ufunc is np.matmul:
            # Call the __matmul__ method
            return self.__matmul__(*inputs)
        elif ufunc is np.add:
            # Call the __add__ method
            return self.__add__(*inputs)
        # Convert all inputs to base class (np.char.chararray) for computation
        inputs = [i.view(np.char.chararray) if isinstance(i, Labels)
                  else i for i in inputs]
        # Perform the ufunc operation
        out = super().__array_ufunc__(ufunc, method, *inputs, **kwargs)
        # Return the result as a Labels object
        return Labels(out)

    def decompose(self) -> list['Labels', ...]:
        """Decompose a Labels object into a list of 1d Labels objects.

        Examples
        --------
        >>> Labels(['a-d', 'a-c', 'b-d', 'b-c']).reshape(2,2).decompose()
        [['a', 'b'], ['d', 'c']]
        >>> Labels(['a-c-e', 'a-c-f', 'a-d-e', 'a-d-f', 'b-c-e', 'b-c-f',
        ... 'b-d-e', 'b-d-f']).reshape(2,2,2).decompose()
        [['a', 'b'], ['c', 'd'], ['e', 'f']]
        >>> (Labels(['a','b','c']) @ Labels(['d','e','f','g'])).reshape(
        ... 2,6).decompose() # doctest: +ELLIPSIS
        [['a-d-a-e-a-f-a-g-b-d-b-e', 'b-f-b-g-c-d-c-e-c-f-c-g'], ['a-d-b-f'...
        """
        new_labels = [[None for _ in range(s)] for s in self.shape]
        for i, dim in enumerate(self.shape):
            for j in range(dim):
                row = np.take(self, j, axis=i).flatten().astype(str)
                common = _longest_common_substring(tuple(map(
                    lambda x: tuple(x.split(self.delimiter, )), row)))
                if len(common) == 0:
                    common = np.unique(row).tolist()
                new_labels[i][j] = self.delimiter.join(common)
            new_labels[i] = _make_array_unique(np.array(new_labels[i]),
                                               self.delimiter)
        return list(map(Labels, new_labels))

    def find(self, value) -> int | tuple[int]:
        """Get the index of the first instance of a value in the Labels"""
        idx = np.where(self == value)[0]
        if (n := len(idx)) == 0:
            if self.delimiter in self[0]:
                splitlist = np.char.split(self, self.delimiter)
                for i in range(len(splitlist[0])):
                    try:
                        return Labels([s[i] for s in splitlist]).find(value)
                    except IndexError:
                        continue
            raise IndexError(f"{value} not found in {self}")
        elif n == 1:
            return int(idx[0])
        else:
            return tuple(map(int, idx))

    def join(self, axis: int = None):
        """Join the labels into a single string using the delimiter

        Parameters
        ----------
        axis : int, optional
            The axis to join along, by default None

        Examples
        --------
        >>> Labels(['a', 'b', 'c']).join()
        'a-b-c'
        >>> Labels(['a', 'b', 'c']).reshape(1,3).join()
        'a-b-c'
        >>> Labels([['a','b'],['c','d']]).join()
        'a-b-c-d'
        >>> Labels([['a','b'],['c','d']]).join(axis=0)
        ['a-b', 'c-d']
        >>> Labels([['a','b'],['c','d']], '').join(axis=1)
        ['ac', 'bd']
        """
        if axis is None:
            return self.delimiter.join(self.flat)
        else:
            labs = self.swapaxes(0, axis)
            return Labels([lab.join() for lab in labs], self.delimiter)


def _make_array_unique(arr: np.ndarray, delimiter: str) -> np.ndarray:
    """Make an array unique by appending a number to duplicate values.

    Parameters
    ----------
    arr : np.ndarray
        The array to make unique.
    delimiter : str
        The delimiter to use when appending a number to duplicate values.

    Returns
    -------
    np.ndarray
        The unique array.

    Examples
    --------
    >>> arr = np.array(['a', 'b', 'c', 'a', 'b', 'c'])
    >>> _make_array_unique(arr, '-')
    array(['a-0', 'b-0', 'c-0', 'a-1', 'b-1', 'c-1'], dtype='<U3')
    >>> arr = np.array(['a', 'b', 'c', 'd', 'e', 'f'])
    >>> _make_array_unique(arr, '-')
    array(['a', 'b', 'c', 'd', 'e', 'f'], dtype='<U1')
    """
    unique, inverse = np.unique(arr, return_inverse=True)
    if len(unique) == len(arr):
        return arr
    counts = np.bincount(inverse)
    max_dtype = np.max([len(u) for u in unique]) + 1 + len(str(max(counts)))
    out = np.empty_like(arr, dtype=f'<U{max_dtype}')
    for i, (u, c) in enumerate(zip(unique, counts)):
        if c == 1:
            out[inverse == i] = u
        else:
            indices = np.where(arr == u)[0]
            for j, index in enumerate(indices):
                out[index] = f"{u}{delimiter}{j}"
    return out


def _longest_common_substring(strings: tuple[tuple[str]]) -> tuple[str]:
    matrix = [[] for _ in range(len(strings))]
    for i in range(len(strings) - 1):
        matrix[i] = _lcs(strings[i], strings[i + 1])
    else:
        matrix[-1] = [True for _ in range(len(strings[-1]))]
    return np.array(strings[0])[np.all(matrix, axis=0)].tolist()


@functools.lru_cache(None)
def _lcs(s1: tuple, s2: tuple) -> list[bool]:
    matrix = [False for _ in range(len(s1))]
    for i in range(len(s1)):
        if s1[i] == s2[i]:
            matrix[i] = True
    return matrix


def is_broadcastable(shp1: tuple[int, ...], shp2: tuple[int, ...]):

    ndim1 = len(shp1)
    ndim2 = len(shp2)
    if ndim1 < ndim2:
        shp1 += (1,) * (ndim2 - ndim1)
    elif ndim2 < ndim1:
        shp2 += (1,) * (ndim1 - ndim2)

    for a, b in zip(shp1, shp2):
        if a == 1 or b == 1 or a == b:
            pass
        else:
            return False
    return True


def get_subset_reorder_indices(array1, array2):
    """Get indices to reorder array1 to match array2"""
    o = [np.where(array1 == i)[0][0] for i in array2 if i in array1]
    return np.array(o)


def add_to_list_if_not_present(lst: list, element: Iterable):
    """Add an element to a list if it is not present. Runs in O(1) time.

    Parameters
    ----------
    lst : list
        The list to add the element to.
    element : Iterable
        The element to add to the list.

    References
    ----------
    [1] https://www.youtube.com/watch?v=PXWL_Xzyrp4

    Examples
    --------
    >>> lst = [1, 2, 3]
    >>> add_to_list_if_not_present(lst, [3, 4, 5])
    >>> lst
    [1, 2, 3, 4, 5]
    """
    seen = set(lst)
    lst.extend(x for x in element if not (x in seen or seen.add(x)))


def inner_all_keys(data: dict, keys: list = None, lvl: int = 0):
    """Get all keys of a nested dictionary.

    Parameters
    ----------
    data : dict
        The nested dictionary to get the keys of.
    keys : list, optional
        The list of keys, by default None
    lvl : int, optional
        The level of the dictionary, by default 0

    Returns
    -------
    tuple
        The tuple of keys.

    Examples
    --------
    >>> data = {'a': {'b': {'c': 1}}}
    >>> inner_all_keys(data)
    (('a',), ('b',), ('c',))
    >>> data = {'a': {'b': {'c': 1}}, 'd': {'b': {'c': 2, 'e': 3}}}
    >>> inner_all_keys(data)
    (('a', 'd'), ('b',), ('c', 'e'))
    """
    if keys is None:
        keys = []
    if isinstance(data, dict):
        if len(keys) < lvl + 1:
            keys.append(list(data.keys()))
        else:
            add_to_list_if_not_present(keys[lvl], data.keys())
        for d in data.values():
            if np.isscalar(d):
                continue
            inner_all_keys(d, keys, lvl + 1)
    elif isinstance(data, np.ndarray):
        data = np.atleast_1d(data)
        rows = range(data.shape[0])
        if len(keys) < lvl + 1:
            keys.append(list(rows))
        else:
            add_to_list_if_not_present(keys[lvl], rows)
        if len(data.shape) > 1:
            if not np.isscalar(data[0]):
                inner_all_keys(data[0], keys, lvl + 1)
    else:
        raise TypeError(f"Unexpected data type: {type(data)}")
    return tuple(map(tuple, keys))


def inner_array(data: dict | np.ndarray) -> np.ndarray | None:
    """Convert a nested dictionary to a nested array.

    Parameters
    ----------
    data : dict or np.ndarray
        The nested dictionary to convert.

    Returns
    -------
    np.ndarray or None
        The converted nested array.

    Examples
    --------
    >>> data = {'a': {'b': {'c': 1}}}
    >>> inner_array(data)
    array([[[1.]]])
    >>> data = {'a': {'b': {'c': 1}}, 'd': {'b': {'c': 2, 'e': 3}}}
    >>> inner_array(data)
    array([[[ 1., nan]],
    <BLANKLINE>
           [[ 2.,  3.]]])
    >>> data = {'a': {'b': {'c': 1, 'd': 2, 'e': 3}, 'f': {'c': 4, 'd': 5}}}
    >>> inner_array(data)
    array([[[ 1.,  2.,  3.],
            [ 4.,  5., nan]]])
    """
    if np.isscalar(data):
        return data
    elif isinstance(data, dict):
        gen_arr = (inner_array(d) for d in data.values())
        arr = [a for a in gen_arr if a is not None]
        if len(arr) > 0:
            return concatenate_arrays(arr, axis=None)

    # Call np.atleast_1d once and store the result in a variable
    data_1d = np.atleast_1d(data)

    # Use the stored result to check the length of data
    if len(data_1d) == 0:
        return
    elif len(data_1d) == 1:
        return data
    else:
        return np.array(data)


def _combine_arrays(*arrays, delim: str = '-') -> np.ndarray:
    # Create a meshgrid of indices
    grids = np.meshgrid(*arrays, indexing='ij')

    # Combine the grids into a single array with string concatenation
    result = np.core.defchararray.add(grids[0], delim)
    for grid in grids[1:]:
        result = np.core.defchararray.add(result, grid)

    return result


def combine(data: dict, levels: tuple[int, int], delim: str = '-') -> dict:
    """Combine any levels of a nested dict into the lower level

    Takes the input nested dict and rearranges the top and bottom
    sub-dictionary.

    Parameters
    ----------
    data: dict
        The nested dict to combine
    levels: tuple[int, int]
        The levels to combine, e.g. (0, 1) will combine the 1st and 2nd level
        of the dict keys into one level at the 2nd level.

    Returns
    -------
    dict
        The combined dict

    Examples
    --------
    >>> data = {'a': {'b': {'c': 1}}}
    >>> combine(data, (0, 2))
    {'b': {'a-c': 1}}
    >>> data = {'a': {'b': {'c': 1}}, 'd': {'b': {'c': 2, 'e': 3}}}
    >>> combine(data, (0, 2))
    {'b': {'a-c': 1, 'd-c': 2, 'd-e': 3}}
    """

    assert levels[0] >= 0, "first level must be >= 0"
    assert levels[1] > levels[0], "second level must be > first level"

    def _combine_helper(data, levels, depth, keys):
        if depth == levels[1]:
            return {f'{keys[levels[0]]}{delim}{k}': v for k, v in data.items()}
        elif depth == levels[0]:
            new_dict = {}
            for k, v in data.items():
                for k2, v2 in _combine_helper(v, levels, depth + 1,
                                              keys + [k]).items():
                    if isinstance(v2, dict):
                        if k2 in new_dict:
                            new_dict[k2] = _merge(new_dict[k2], v2)
                        else:
                            new_dict[k2] = v2
                    else:
                        new_dict[k2] = v2
            return new_dict
        else:
            return {k: _combine_helper(v, levels, depth + 1, keys + [k]) for
                    k, v in data.items()}

    def _merge(d1: dict, d2: dict) -> dict:
        for k, v in d2.items():
            if isinstance(v, dict):
                d1[k] = _merge(d1.get(k, {}), v)
            else:
                d1[k] = v
        return d1

    result = _combine_helper(data, levels, 0, [])

    return result


def stack_la(arrays: tuple[LabeledArray, ...], new_labels: list[str, ...]
             ) -> LabeledArray:
    """Stack a sequence of LabeledArrays along a new axis.

    Parameters
    ----------
    arrays : LabeledArray
        The LabeledArrays to stack.
    new_labels : Labels
        The new labels for the stacked axis.

    Returns
    -------
    LabeledArray
        The stacked LabeledArray.

    Examples
    --------
    >>> arr1 = LabeledArray([[1, 2],[3, 4]], labels=[('a', 'b'), ('c', 'd')])
    >>> arr2 = LabeledArray([[5, 6, 7],[7, 8, 9]],
    ... labels=[('a', 'b'), ('c', 'd', 'e')])
    >>> stack_la((arr1, arr2), ['1', '2'])
    array([[[ 1.,  2., nan],
            [ 3.,  4., nan]],
    <BLANKLINE>
           [[ 5.,  6.,  7.],
            [ 7.,  8.,  9.]]])
    labels(['1', '2']
           ['a', 'b']
           ['c', 'd', 'e'])
    """
    new_array = concatenate_arrays([a.__array__() for a in arrays], None)

    # get the longest labels in each axis
    new_labels = [Labels(new_labels)]
    for i in range(new_array.ndim - 1):
        new_labels.append(max((a.labels[i] for a in arrays), key=len))

    return LabeledArray(new_array, new_labels)


def get_elbow(data: np.ndarray) -> int:
    """Draws a line between the first and last points in a dataset and finds
    the point furthest from that line.

    Parameters
    ----------
    data : array
        The data to find the elbow in.

    Returns
    -------
    int
        The index of the elbow point.

    Examples
    --------
    >>> data = np.array([0, 1, 2, 3, 4, 4.5, 5, 5.5, 6, 7, 8, 9, 10])
    >>> get_elbow(data)
    4
    >>> data = np.array([1, 2, 3, 4, 5, 4.5, 4, 3.5, 3, 2, 1])
    >>> get_elbow(data)
    4
    """
    nPoints = len(data)
    allCoord = np.vstack((range(nPoints), data)).T
    np.array([range(nPoints), data])
    firstPoint = allCoord[0]
    lineVec = allCoord[-1] - allCoord[0]
    lineVecNorm = lineVec / np.sqrt(np.sum(lineVec ** 2))
    vecFromFirst = allCoord - firstPoint
    scalarProduct = np.sum(vecFromFirst * repmat(
        lineVecNorm, nPoints, 1), axis=1)
    vecFromFirstParallel = np.outer(scalarProduct, lineVecNorm)
    vecToLine = vecFromFirst - vecFromFirstParallel
    distToLine = np.sqrt(np.sum(vecToLine ** 2, axis=1))
    # set distance to points below lineVec to 0
    distToLine[vecToLine[:, 1] < 0] = 0
    idxOfBestPoint = np.argmax(distToLine)
    return idxOfBestPoint


def events_in_order(inst: mne.BaseEpochs) -> list[str]:
    ids = {v: k for k, v in inst.event_id.items()}
    return [ids[e[2]] for e in inst.events]


if __name__ == "__main__":
    import os
    from ieeg.io import get_data
    import mne

    conds = {"resp": ((-1, 1), "Response/LS"), "aud_ls": ((-0.5, 1.5),
                                                          "Audio/LS"),
             "aud_lm": ((-0.5, 1.5), "Audio/LM"), "aud_jl": ((-0.5, 1.5),
                                                             "Audio/JL"),
             "go_ls": ((-0.5, 1.5), "Go/LS"), "go_lm": ((-0.5, 1.5), "Go/LM"),
             "go_jl": ((-0.5, 1.5), "Go/JL")}
    task = "SentenceRep"
    root = os.path.expanduser("~/Box/CoganLab")
    # layout = get_data(task, root=root)
    folder = 'stats_old'
    mne.set_log_level("ERROR")

    arr = np.arange(24).reshape((2, 3, 4))
    labels = (('a', 'b'), ('c', 'd', 'e'), ('f', 'g', 'h', 'i'))
    ad = LabeledArray(arr, labels)
    Labels(['a', 'b', 'c']) @ Labels(['d', 'e', 'f'])

    labels = Labels(np.arange(1000))
    l2d = labels @ labels
    x = l2d.reshape((10, -1)).decompose()
    x = np.moveaxis(ad, 0, 1)


def _cat_test():
    """Test concatenation of arrays

    Concatenate a list of arrays along a given axis.

    Parameters
    ----------
    arrays
    axis

    Examples
    --------
    >>> import numpy as np
    >>> a = np.array([[1, 2, 3]])
    >>> b = np.array([[4, 5]])
    >>> c = np.array([[6, 7, 8, 9]])
    >>> concatenate_arrays([a, b, c])
    array([[ 1.,  2.,  3., nan],
           [ 4.,  5., nan, nan],
           [ 6.,  7.,  8.,  9.]])
    >>> concatenate_arrays([a, b, c], axis=1)
    array([[1., 2., 3., 4., 5., 6., 7., 8., 9.]])
    """
